#!/usr/bin/env python
"""
_MarkDatasetSubscribed_

MySQL implementation of PhEDExInjector.Database.MarkDatasetSubscribed
"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class MarkDatasetSubscribed(DBFormatter):
    """
    _MarkDatasetSubscribed_

    Marks the given dataset subscription as subscribed in the database
    """

    sql = "UPDATE dbsbuffer_dataset_subscription SET subscribed = 1 WHERE id = :id"

    def execute(self, subIds, conn=None, transaction=False):
        binds = []
        for subId in subIds:
            binds.append({"id": subId})
        self.dbi.processData(self.sql, binds,
                             conn=conn, transaction=transaction)
        return



#!/usr/bin/env python

"""
Oracle implementation of GetChecksum
"""





from WMCore.WMBS.MySQL.Files.GetChecksum import GetChecksum as MySQLGetChecksum

class GetChecksum(MySQLGetChecksum):
    """
    Identical to MySQL

    """



#!/usr/bin/env python
"""
_BuildTools_

Utils to assist in the build process


"""

from builtins import object
from future.utils import viewvalues

import os
import logging
from WMCore.WMSpec.ConfigSectionTree import nodeName


class File(object):
    """
    _File_

    Class representing a file added to a directory.

    Contains a source for the file and a name within
    the directory

    """
    def __init__(self, directory, name, source):
        self.directory = directory
        self.name = name
        self.source = source




    def path(self):
        """
        _path_

        Get name of this file within directory structure

        """
        return "%s/%s" % (self.directory.path(), self.name)

    def fetch(self, targetDir):
        """
        _fetch_

        Get the source and put it in the target dir.

        Note: for now this uses cp, could use other
        things based on source type, eg http:// etc etc

        """
        command = "/bin/cp -rf %s %s/%s" % (self.source,
                                        targetDir,
                                        self.name)
        logging.info("fetch:%s" % command)
        os.system(command)
        return




class Directory(object):
    """
    _Directory_

    structure representing a dir, to which files can be attached


    """
    def __init__(self, name):
        self.name = name
        self.parent = None
        self.children = {}
        self.files = {}
        self.physicalPath = None


    def addDirectory(self, name):
        """
        _addDirectory_

        Add a new child Directory to this.
        Return reference to new Directory instance

        """
        if name in self.children:
            return self.children[name]

        self.children[name] = Directory(name)
        self.children[name].parent = self
        return self.children[name]


    def addFile(self, source, targetName = None):
        """
        _addFile_

        Add a file to this directory.
        The file will be pulled in from the source specified.
        targetName is the optional name of the file in this
        directory. If not specified, the basename of the file
        will be used

        """
        target = targetName
        if target == None:
            target = os.path.basename(source)

        if target in self.files:
            msg = "File %s already exists in directory %s" % (
                self.name, target)
            raise RuntimeError(msg)

        newFile = File(self, target, source)
        self.files[target] = newFile
        return


    def path(self):
        """
        _path_

        Get name of this dir within directory structure

        """
        if self.parent == None:
            return self.name
        return "%s/%s" % (self.parent.path(), self.name)


    def create(self, targetDir):
        """
        _create_

        Make this directory in the targetDirectory provided,
        pull in all files and then recursively create any
        children

        """
        newDir = "%s/%s" % (targetDir, self.name)
        logging.info("create(%s)" % newDir)
        if not os.path.exists(newDir):
            os.makedirs(newDir)
        for f in viewvalues(self.files):
            f.fetch(newDir)

        for child in viewvalues(self.children):
            child.create(newDir)
        return

    def __str__(self):
        result = "%s\n" % self.path()
        for f in viewvalues(self.files):
            result += "%s ==> %s\n" % (f.path(), f.source)
        for d in viewvalues(self.children):
            result += str(d)

        return result








def processDir( cfgSect, parent):
    """
    _processDir_

    Process a ConfigSection based directory

    """
    for setting in cfgSect._internal_settings:
        value = getattr(cfgSect, setting)
        if not isinstance(value, dict): continue
        parent.addFile(value['Source'], value['Target'])

    for subdir in cfgSect._internal_children:
        newsubdir = parent.addDirectory(subdir)
        processDir(getattr(cfgSect, subdir), newsubdir)
    return


def makeDirectory(step):
    """
    _makeDirectory_

    Create a Directory & file structure from the step provided

    """

    dirs = getattr(step.build.directories, nodeName(step))
    topDir = Directory(nodeName(step))

    processDir(dirs, topDir)

    return topDir



"""
_GetHeartbeatInfo_

MySQL implementation of GetHeartbeatInfo
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class GetAllHeartbeatInfo(DBFormatter):

    sql = """SELECT comp.name as name, comp.pid, worker.name as worker_name,
                 worker.state, worker.last_updated, comp.update_threshold,
                 worker.poll_interval, worker.cycle_time, worker.outcome,
                 worker.last_error, worker.error_message
             FROM wm_workers worker
             INNER JOIN wm_components comp ON comp.id = worker.component_id
             """

    def execute(self, conn = None, transaction = False):

        result = self.dbi.processData(self.sql, conn = conn,
                             transaction = transaction)
        return self.formatDict(result)



"""
_Create_DBS3Buffer_

Implementation of Create_DBSBuffer for MySQL.
"""

import threading

from WMCore.Database.DBCreator import DBCreator


class Create(DBCreator):
    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        myThread = threading.currentThread()

        DBCreator.__init__(self, myThread.logger, myThread.dbi)

        #
        # Tables, functions, procedures and sequences
        #
        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_dataset (
                 id              INTEGER      AUTO_INCREMENT,
                 path            VARCHAR(500) COLLATE latin1_general_cs NOT NULL,
                 processing_ver  INTEGER      NOT NULL,
                 acquisition_era VARCHAR(255),
                 valid_status    VARCHAR(20),
                 global_tag      VARCHAR(255),
                 parent          VARCHAR(500),
                 prep_id         VARCHAR(255),
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_dat UNIQUE (path))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_dataset_subscription (
                 id                     INTEGER      AUTO_INCREMENT,
                 dataset_id             INTEGER      NOT NULL,
                 site                   VARCHAR(100) NOT NULL,
                 custodial              INTEGER      DEFAULT 0,
                 priority               VARCHAR(10)  DEFAULT 'Low',
                 subscribed             INTEGER      DEFAULT 0,
                 delete_blocks          INTEGER,
                 dataset_lifetime       INTEGER      DEFAULT 0 NOT NULL,
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_dat_sub UNIQUE (dataset_id, site, custodial, priority))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_algo (
                 id             INTEGER       AUTO_INCREMENT,
                 app_name       VARCHAR(100),
                 app_ver        VARCHAR(100),
                 app_fam        VARCHAR(100),
                 pset_hash      VARCHAR(700),
                 config_content LONGTEXT,
                 in_dbs         INTEGER,
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_alg UNIQUE (app_name, app_ver, app_fam, pset_hash))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_algo_dataset_assoc (
                 id         INTEGER AUTO_INCREMENT,
                 algo_id    INTEGER NOT NULL,
                 dataset_id INTEGER NOT NULL,
                 in_dbs     INTEGER DEFAULT 0,
                 PRIMARY KEY (id))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_workflow (
                 id                           INTEGER       AUTO_INCREMENT,
                 name                         VARCHAR(700),
                 task                         VARCHAR(700),
                 block_close_max_wait_time    INTEGER,
                 block_close_max_files        INTEGER,
                 block_close_max_events       INTEGER,
                 block_close_max_size         BIGINT,
                 completed                    INTEGER       DEFAULT 0,
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_wor UNIQUE (name, task))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_file (
                 id                    INTEGER      AUTO_INCREMENT,
                 lfn                   VARCHAR(500) NOT NULL,
                 filesize              BIGINT,
                 events                BIGINT UNSIGNED,
                 dataset_algo          INTEGER      NOT NULL,
                 block_id              INTEGER,
                 status                VARCHAR(20),
                 in_phedex             INTEGER      DEFAULT 0,
                 workflow              INTEGER,
                 LastModificationDate  INTEGER,
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_fil UNIQUE (lfn))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_file_parent (
                 child  INTEGER NOT NULL,
                 parent INTEGER NOT NULL,
                 CONSTRAINT pk_dbs_fil_par PRIMARY KEY (child, parent))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_file_runlumi_map (
                 filename    INTEGER NOT NULL,
                 run         INTEGER NOT NULL,
                 lumi        INTEGER NOT NULL,
                 num_events  BIGINT UNSIGNED)"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_location (
                 id       INTEGER      AUTO_INCREMENT,
                 pnn  VARCHAR(255) NOT NULL,
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_loc UNIQUE (pnn))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_file_location (
                 filename INTEGER NOT NULL,
                 location INTEGER NOT NULL,
                 CONSTRAINT pk_dbs_fil_loc PRIMARY KEY (filename, location))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_block (
                 id           INTEGER      AUTO_INCREMENT,
                 dataset_id   INTEGER      NOT NULL,
                 blockname    VARCHAR(250) NOT NULL,
                 location     INTEGER      NOT NULL,
                 create_time  INTEGER,
                 status       VARCHAR(20),
                 deleted      INTEGER      DEFAULT 0,
                 rule_id      VARCHAR(40)  NOT NULL DEFAULT '0',
                 PRIMARY KEY (id),
                 CONSTRAINT uq_dbs_blo UNIQUE (blockname, location))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_checksum_type (
                 id   INTEGER      AUTO_INCREMENT,
                 type VARCHAR(255),
                 PRIMARY KEY (id))"""

        self.create[len(self.create)] = \
            """CREATE TABLE dbsbuffer_file_checksums (
                 fileid  INTEGER,
                 typeid  INTEGER,
                 cksum   VARCHAR(100),
                 CONSTRAINT pk_dbs_fil_che PRIMARY KEY (fileid, typeid))"""

        #
        # Indexes
        #
        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_che_1 ON dbsbuffer_file_checksums (typeid)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_che_2 ON dbsbuffer_file_checksums (fileid)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_dat_sub_1 ON dbsbuffer_dataset_subscription (dataset_id)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_par_1 ON dbsbuffer_file_parent (child)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_par_2 ON dbsbuffer_file_parent (parent)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_blo_1 ON dbsbuffer_block (location)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_blo_2 ON dbsbuffer_block (dataset_id)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_alg_ass_1 ON dbsbuffer_algo_dataset_assoc (dataset_id)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_alg_ass_2 ON dbsbuffer_algo_dataset_assoc (algo_id)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_run_1 ON dbsbuffer_file_runlumi_map (filename)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_loc_1 ON dbsbuffer_file_location (location)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_loc_2 ON dbsbuffer_file_location (filename)"""

        self.indexes[len(self.indexes)] = \
            """CREATE INDEX idx_dbs_fil_1 ON dbsbuffer_file (workflow)"""

        #
        # Constraints
        #
        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_checksums
                 ADD CONSTRAINT fk_file_checksums_typeid
                 FOREIGN KEY (typeid)
                 REFERENCES dbsbuffer_checksum_type(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_checksums
                 ADD CONSTRAINT fk_file_checksums_fileid
                 FOREIGN KEY (fileid)
                 REFERENCES dbsbuffer_file(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_dataset_subscription
                 ADD CONSTRAINT fk_dsetsubscription_datasetid
                 FOREIGN KEY (dataset_id)
                 REFERENCES dbsbuffer_dataset(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_parent
                 ADD CONSTRAINT fk_file_parent_child
                 FOREIGN KEY (child)
                 REFERENCES dbsbuffer_file(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_parent
                 ADD CONSTRAINT fk_file_parent_parent
                 FOREIGN KEY (parent)
                 REFERENCES dbsbuffer_file(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_block
                 ADD CONSTRAINT fk_block_location
                 FOREIGN KEY (location)
                 REFERENCES dbsbuffer_location(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_block
                 ADD CONSTRAINT fk_block_dataset_id
                 FOREIGN KEY (dataset_id)
                 REFERENCES dbsbuffer_dataset(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_algo_dataset_assoc
                 ADD CONSTRAINT fk_algodset_assoc_dataset_id
                 FOREIGN KEY (dataset_id)
                 REFERENCES dbsbuffer_dataset(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_algo_dataset_assoc
                 ADD CONSTRAINT fk_algodset_assoc_algo_id
                 FOREIGN KEY (algo_id)
                 REFERENCES dbsbuffer_algo(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_runlumi_map
                 ADD CONSTRAINT fk_file_runlumi_filename
                 FOREIGN KEY (filename)
                 REFERENCES dbsbuffer_file(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_location
                 ADD CONSTRAINT fk_file_location_location
                 FOREIGN KEY (location)
                 REFERENCES dbsbuffer_location(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file_location
                 ADD CONSTRAINT fk_file_location_filename
                 FOREIGN KEY (filename)
                 REFERENCES dbsbuffer_file(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file
                 ADD CONSTRAINT fk_file_workflow
                 FOREIGN KEY (workflow)
                 REFERENCES dbsbuffer_workflow(id)
                 ON DELETE CASCADE"""

        self.constraints[len(self.constraints)] = \
            """ALTER TABLE dbsbuffer_file
                 ADD CONSTRAINT fk_file_dataset_algo
                 FOREIGN KEY (dataset_algo)
                 REFERENCES dbsbuffer_algo_dataset_assoc(id)
                 ON DELETE CASCADE"""

        checksumTypes = ['cksum', 'adler32', 'md5']
        for i in checksumTypes:
            checksumTypeQuery = """INSERT INTO dbsbuffer_checksum_type
                                   (type)
                                   VALUES ('%s')
                                   """ % (i)
            self.inserts["wmbs_checksum_type_%s" % (i)] = checksumTypeQuery

        for i in self.create:
            self.create[i] += " ENGINE=InnoDB ROW_FORMAT=DYNAMIC"



#!/usr/bin/env python
"""
_HeritageLFNParent_

MySQL implementation of DBSBufferFiles.HeritageLFNParent
"""

from WMCore.Database.DBFormatter import DBFormatter

class BulkHeritageParent(DBFormatter):
    sql = """INSERT INTO dbsbuffer_file_parent (child, parent) VALUES
               ((SELECT id FROM dbsbuffer_file WHERE lfn = :child),
               (SELECT id FROM dbsbuffer_file WHERE lfn = :parent)) """

    def execute(self, binds, conn = None, transaction = False):
        """
        _execute_

        This requires you to send in a list of binds in the form:
        {child, parent}
        """
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_FailFiles_

Oracle implementation of Subscription.FailFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.FailFiles import FailFiles \
     as FailFilesMySQL

class FailFiles(FailFilesMySQL):
    sql = """INSERT INTO wmbs_sub_files_failed (subscription, fileid)
               SELECT :subscription, :fileid FROM dual WHERE NOT EXISTS
                 (SELECT fileid FROM wmbs_sub_files_failed
                    WHERE fileid = :fileid AND subscription = :subscription)"""



#!/usr/bin/env python
# pylint: disable=W0201, W0102
# W0201: Steve defines all global vars in __call__
#   I don't know why, but I'm not getting blamed for it
# W0102: Dangerous default values?  I live on danger!
#   Allows us to use a dict as a default
"""
_StoreResults_

Standard StoreResults workflow.
"""

from Utils.Utilities import makeList, makeNonEmptyList
from WMCore.Lexicon import dataset, block, physicsgroup, cmsname
from WMCore.WMSpec.StdSpecs.StdBase import StdBase


class StoreResultsWorkloadFactory(StdBase):
    """
    _StoreResultsWorkloadFactory_

    Stamp out StoreResults workloads.
    """

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a StoreResults workload with the given parameters.
        """
        # first of all, we update the merged LFN based on the physics group
        arguments['MergedLFNBase'] += "/" + arguments['PhysicsGroup'].lower()
        StdBase.__call__(self, workloadName, arguments)

        (inputPrimaryDataset, inputProcessedDataset, inputDataTier) = self.inputDataset[1:].split("/")

        workload = self.createWorkload()

        mergeTask = workload.newTask("StoreResults")
        self.addRuntimeMonitors(mergeTask)
        mergeTaskCmssw = mergeTask.makeStep("cmsRun1")
        mergeTaskCmssw.setStepType("CMSSW")

        mergeTaskStageOut = mergeTaskCmssw.addStep("stageOut1")
        mergeTaskStageOut.setStepType("StageOut")

        mergeTaskLogArch = mergeTaskCmssw.addStep("logArch1")
        mergeTaskLogArch.setStepType("LogArchive")

        self.addLogCollectTask(mergeTask, taskName="StoreResultsLogCollect")

        mergeTask.setTaskType("Merge")
        mergeTask.applyTemplates()

        mergeTask.addInputDataset(name=self.inputDataset,
                                  primary=inputPrimaryDataset,
                                  processed=inputProcessedDataset,
                                  tier=inputDataTier,
                                  dbsurl=self.dbsUrl,
                                  block_blacklist=self.blockBlacklist,
                                  block_whitelist=self.blockWhitelist,
                                  run_blacklist=self.runBlacklist,
                                  run_whitelist=self.runWhitelist)

        splitAlgo = "ParentlessMergeBySize"
        mergeTask.setSplittingAlgorithm(splitAlgo,
                                        max_merge_size=self.maxMergeSize,
                                        min_merge_size=self.minMergeSize,
                                        max_merge_events=self.maxMergeEvents)

        mergeTaskCmsswHelper = mergeTaskCmssw.getTypeHelper()
        mergeTaskCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                        scramArch=self.scramArch)
        mergeTaskCmsswHelper.setGlobalTag(self.globalTag)
        mergeTaskCmsswHelper.setSkipBadFiles(True)
        mergeTaskCmsswHelper.setDataProcessingConfig("do_not_use", "merge")

        self.addOutputModule(mergeTask, "Merged",
                             primaryDataset=inputPrimaryDataset,
                             dataTier=self.dataTier,
                             filterName=None,
                             forceMerged=True)

        workload.setLFNBase(self.mergedLFNBase, self.unmergedLFNBase)
        workload.setDashboardActivity("StoreResults")

        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        workload.setTaskPropertiesFromWorkload()

        return workload

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = StdBase.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "StoreResults", "optional": False},
                    "InputDataset": {"optional": False, "validate": dataset, "null": False},
                    "ConfigCacheID": {"optional": True, "null": True},
                    "DataTier": {"default": "USER", "type": str,
                                 "optional": True, "validate": None,
                                 "attr": "dataTier", "null": False},
                    "PhysicsGroup": {"default": "", "optional": False,
                                     "null": False, "validate": physicsgroup},
                    "MergedLFNBase": {"default": "/store/results", "type": str,
                                      "optional": True, "validate": None,
                                      "attr": "mergedLFNBase", "null": False},
                    # site whitelist shouldn't be allowed, but let's make an exception for StoreResults
                    "SiteWhitelist": {"default": [], "type": makeNonEmptyList, "assign_optional": False,
                                      "validate": lambda x: all([cmsname(y) for y in x])},
                    "BlockBlacklist": {"default": [], "type": makeList,
                                       "optional": True, "validate": lambda x: all([block(y) for y in x]),
                                       "attr": "blockBlacklist", "null": False},
                    "BlockWhitelist": {"default": [], "type": makeList,
                                       "optional": True, "validate": lambda x: all([block(y) for y in x]),
                                       "attr": "blockWhitelist", "null": False},
                    "RunBlacklist": {"default": [], "type": makeList,
                                     "optional": True, "validate": lambda x: all([int(y) > 0 for y in x]),
                                     "attr": "runBlacklist", "null": False},
                    "RunWhitelist": {"default": [], "type": makeList,
                                     "optional": True, "validate": lambda x: all([int(y) > 0 for y in x]),
                                     "attr": "runWhitelist", "null": False}}
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs



#!/usr/bin/env python

"""
MySQL implementation of AddCKType
"""





from WMCore.Database.DBFormatter import DBFormatter

class AddCKType(DBFormatter):
    sql = """INSERT INTO wmbs_checksum_type (type)
               VALUES (:cktype)"""

    def execute(self, cktype = None, conn = None, transaction = False):

        binds = {'cktype': cktype}

        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_LogArchive_

Basic Emulator for LogArchive Step

"""
from __future__ import print_function

import os
import os.path
import re
import shutil

from WMCore.WMSpec.Steps.Emulator import Emulator

class LogArchive(Emulator):
    """
    _LogArchive_

    Emulate the execution of a LogArchive Step

    """
    def pre(self):
        """
        _pre_

        Emulate the stage out pre step

        """
        return None

    def execute(self):
        """
        _execute_

        Emulate LogArchive execution

        """
        self.step.section_("execution")
        self.step.execution.exitStatus = 0
        self.step.section_("emulation")
        self.step.emulation.emulatedBy = str(self.__class__.__name__)

        print("Emulating LogArchive Step")

        #Find all the reports produced in previous steps
        listOfReports = []
        taskLocation = self.stepSpace.taskSpace.location
        task = self.stepSpace.getWMTask()
        for stepName in task.listAllStepNames():
            #First go through all the steps
            step = task.getStep(stepName)
            if step.stepType() == "CMSSW":
                #If we have a CMSSW step, find the report
                stepPath = os.path.join(taskLocation, stepName)
                if 'Report.pkl' in os.listdir(stepPath):
                    shutil.copy(os.path.join(stepPath, 'Report.pkl'), taskLocation)
                    #If we have the report, copy it and exit
                    return
        for dir in os.listdir(taskLocation):
            if not os.path.isdir(os.path.join(taskLocation, dir)):
                continue
            if 'Report.pkl' in os.listdir(os.path.join(taskLocation, dir)):
                listOfReports.append(os.path.join(taskLocation, dir, 'Report.pkl'))

        #For now just move the first Report to the top directory
        if len(listOfReports) > 0:
            shutil.copy(listOfReports[0], taskLocation)


        return



    def post(self):
        """
        _post_

        Emulate post stage out

        """
        return None



#!/usr/bin/env python
"""
_ListFilesetByTask_

Oracle implementation of Fileset.ListFilesetByTask
"""





from WMCore.WMBS.MySQL.Fileset.ListFilesetByTask import ListFilesetByTask as ListFilesetByTaskMySQL

class ListFilesetByTask(ListFilesetByTaskMySQL):
    pass



#!/usr/bin/env python
"""
_LoadFromIDWithType_

MySQL implementation of LoadFromIDWithType

Created on Oct 3, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.LoadFromID import LoadFromID

class LoadFromIDWithType(LoadFromID):
    """
    _LoadFromIDWithType_

    Load jobs by ID but include the type
    of the subscription they belong to
    """

    sql = """SELECT wmbs_job.id, jobgroup, wmbs_job.name AS name,
                    wmbs_job_state.name AS state, wmbs_job.state_time, retry_count,
                    couch_record,  cache_dir, wmbs_location.site_name AS location,
                    outcome AS bool_outcome, fwjr_path AS fwjr_path,
                    wmbs_sub_types.name AS type
             FROM wmbs_job
               LEFT OUTER JOIN wmbs_location ON
                 wmbs_job.location = wmbs_location.id
               LEFT OUTER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
             INNER JOIN wmbs_jobgroup ON wmbs_jobgroup.id = wmbs_job.jobgroup
             INNER JOIN wmbs_subscription ON wmbs_subscription.id = wmbs_jobgroup.subscription
             INNER JOIN wmbs_sub_types ON wmbs_sub_types.id = wmbs_subscription.subtype
             WHERE wmbs_job.id = :jobid"""



#!/usr/bin/env python
"""
_BuildMaster_

Overseer object that traverses a task and invokes the type based builder
for each step

"""




from builtins import object
import os

import WMCore.WMSpec.Steps.StepFactory as StepFactory


header = \
"""
#!/usr/bin/env python

#
# WMTaskSpace entry point
# Autogenerated by WMCore.WMSpec.Steps.BuildMaster Module
__all__ = []

from WMCore.WMRuntime.Bootstrap import establishTaskSpace
def _Locator():
    pass
args = {}

"""


def initialiseWMTaskSpace(directory, taskName):
    """
    _initialiseWMTaskSpace_

    Build a top level working directory with the appropriate __init__.py
    module to define the top of a working area for a job

    """
    if not os.path.exists(directory):
        msg = "Directory %s not found" % directory
        raise RuntimeError(msg)

    spaceDir = "%s/WMTaskSpace" % directory
    if os.path.exists(spaceDir):
        os.system("/bin/rm -r %s" % spaceDir)
    os.makedirs(spaceDir)

    initFile = "%s/__init__.py" % spaceDir
    with open(initFile, "w") as handle:
        handle.write(header)
        handle.write("""args["TaskName"] = "%s"\n""" % taskName)
        handle.write("""args["Locator"] = _Locator\n""")
        handle.write("""taskSpace = establishTaskSpace(**args)\n""")
    return spaceDir



class BuildMaster(object):
    """
    _BuildMaster_

    """
    def __init__(self, workingDir):
        self.workDir = workingDir
        self.taskSpace = None


    def __call__(self, task):
        """
        _operator(task)_

        Invoke the builder on the task provided

        TODO: Exception handling

        """
        taskName = task.getPathName()
        self.taskSpace = initialiseWMTaskSpace(self.workDir, taskName)
        for step in task.steps().nodeIterator():
            stepType = step.stepType
            builder = StepFactory.getStepBuilder(stepType)
            builder(step, task.name(), self.taskSpace)



#!/usr/bin/env python
"""
_ListThresholds_

This module list thresholds for the given sites for Oracle
"""


from WMCore.ResourceControl.MySQL.ListThresholds import ListThresholds \
  as ListThresholdsMySQL
class ListThresholds(ListThresholdsMySQL):
    pass



#!/usr/bin/env python
# pylint: disable=W0102, C0103, W0621, W0105, W0703
"""
_ProcessPool_

"""
from __future__ import print_function

from builtins import range, object

import zmq
import subprocess
import sys
import logging
import os
import threading
import traceback
import pickle

from Utils.PythonVersion import PY3

from logging.handlers import RotatingFileHandler

from WMCore.WMFactory import WMFactory
from WMCore.WMInit import WMInit

from WMCore.Services.Requests import JSONRequests

from WMCore.Agent.HeartbeatAPI import HeartbeatAPI

from WMCore.WMException import WMException


class ProcessPoolException(WMException):
    """
    _ProcessPoolException_

    Raise some exceptions
    """


class ProcessPoolWorker(object):
    """
    _ProcessPoolWorker_

    A basic worker object
    """

    def __init__(self, config):
        self.config = config

        return

    def __call__(self, input):
        """
        __call__

        input should be a single piece of work - a single list, a single
        dictionary, or a single variable of some sort
        """

        return


class ProcessPool(object):
    def __init__(self, slaveClassName, totalSlaves, componentDir,
                 config, namespace='WMComponent', inPort='5555',
                 outPort='5558'):
        """
        __init__

        Constructor for the process pool.  The slave class name must be based
        inside the WMComponent namespace.  For examples, the JobAccountant would
        pass in 'JobAccountant.AccountantWorker' to run the AccountantWorker
        class.  All log files will be stored in the component directory that is
        passed in.  Each slave will have its own log file.

        Note that the config is only used to determine database connection
        parameters.  It is not passed to the slave class.  The slaveInit
        parameter will be serialized and passed to the slave class's
        constructor.
        """
        self.enqueueIndex = 0
        self.dequeueIndex = 0
        self.runningWork = 0

        # Use the Services.Requests JSONizer, which handles __to_json__ calls
        self.jsonHandler = JSONRequests()

        # heartbeat should be registered at this point
        if getattr(config.Agent, "useHeartbeat", True):
            self.heartbeatAPI = HeartbeatAPI(getattr(config.Agent, "componentName", "ProcPoolSlave"))

        self.slaveClassName = slaveClassName
        self.componentDir = componentDir
        self.config = config
        self.versionString = "python3" if PY3 else "python2"

        self.workers = []
        self.nSlaves = totalSlaves
        self.namespace = namespace
        self.inPort = inPort
        self.outPort = outPort


        # Pickle the config
        self.configPath = os.path.join(componentDir, '%s_config.pkl' % slaveClassName)
        if os.path.exists(self.configPath):
            # Then we note it and overwrite it
            msg = "Something's in the way of the ProcessPool config: %s" % self.configPath
            logging.error(msg)
        with open(self.configPath, 'wb') as f:
            pickle.dump(config, f)

        # Set up ZMQ
        try:
            context = zmq.Context()
            self.sender = context.socket(zmq.PUSH)
            self.sender.bind("tcp://*:%s" % inPort)
            self.sink = context.socket(zmq.PULL)
            self.sink.bind("tcp://*:%s" % outPort)
        except zmq.ZMQError:
            # Try this again in a moment to see
            # if it's just being held by something pre-existing
            import time
            time.sleep(1)
            logging.error("Blocked socket on startup: Attempting sleep to give it time to clear.")
            try:
                context = zmq.Context()
                self.sender = context.socket(zmq.PUSH)
                self.sender.bind("tcp://*:%s" % inPort)
                self.sink = context.socket(zmq.PULL)
                self.sink.bind("tcp://*:%s" % outPort)
            except Exception as ex:
                msg = "Error attempting to open TCP sockets\n"
                msg += str(ex)
                logging.error(msg)
                import traceback
                print(traceback.format_exc())
                raise ProcessPoolException(msg)

        # Now actually create the slaves
        self.createSlaves()

        return

    def createSlaves(self):
        """
        _createSlaves_

        Create the slaves by using the values from __init__()
        Moving it into a separate function allows us to restart
        all of them.
        """

        totalSlaves = self.nSlaves
        slaveClassName = self.slaveClassName
        config = self.config
        namespace = self.namespace
        inPort = self.inPort
        outPort = self.outPort

        slaveArgs = [self.versionString, __file__, self.slaveClassName, inPort,
                     outPort, self.configPath, self.componentDir, self.namespace]

        count = 0
        while totalSlaves > 0:
            # For each worker you want create a slave process
            # That process calls this code (WMCore.ProcessPool) and opens
            # A process pool that loads the designated class
            slaveProcess = subprocess.Popen(slaveArgs, stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE)
            self.workers.append(slaveProcess)
            totalSlaves -= 1
            count += 1

        return

    def _subProcessName(self, slaveClassName, sequence):
        """ subProcessName for heartbeat
            could change to use process ID as a suffix
        """
        return "%s_%s" % (slaveClassName, sequence + 1)

    def __del__(self):
        """
        __del__

        Kill all the workers processes by sending them an invalid JSON object.
        This will cause them to shut down.
        """
        self.close()
        return

    def close(self):
        """
        _close_

        Close shuts down all the active systems by:

        a) Sending STOP commands for all workers
        b) Closing the pipes
        c) Shutting down the workers themselves
        """
        for i in range(self.nSlaves):
            try:
                encodedWork = self.jsonHandler.encode('STOP')
                if PY3:
                    self.sender.send_string(encodedWork)
                else:
                    self.sender.send(encodedWork)
            except Exception as ex:
                # Might be already failed.  Nothing you can
                # really do about that.
                logging.error("Failure killing running process: %s" % str(ex))
                pass

        try:
            self.sender.close()
        except:
            # We can't really do anything if we fail
            pass
        try:
            self.sink.close()
        except:
            # We can't do anything if we fail
            pass

        # Now close the workers by hand
        for worker in self.workers:
            try:
                worker.join()
            except Exception as ex:
                try:
                    worker.terminate()
                except Exception as ex2:
                    logging.error("Failure to join or terminate process")
                    logging.error(str(ex))
                    logging.error(str(ex2))
                    continue
        self.workers = []
        return

    def enqueue(self, work, list=False):
        """
        __enqeue__

        Assign work to the workers processes.  The work parameters must be a
        list where each item in the list can be serialized into JSON.

        If list is True, the entire list is sent as one piece of work
        """
        if len(self.workers) < 1:
            # Someone's shut down the system
            msg = "Attempting to send work after system failure and shutdown!\n"
            logging.error(msg)
            raise ProcessPoolException(msg)

        if not list:
            for w in work:
                encodedWork = self.jsonHandler.encode(w)
                if PY3:
                    self.sender.send_string(encodedWork)
                else:
                    self.sender.send(encodedWork)
                self.runningWork += 1
        else:
            encodedWork = self.jsonHandler.encode(work)
            if PY3:
                self.sender.send_string(encodedWork)
            else:
                self.sender.send(encodedWork)
            self.runningWork += 1

        return

    def dequeue(self, totalItems=1):
        """
        __dequeue__

        Retrieve completed work from the slave workers.  This method will block
        until enough work has been completed.
        """
        completedWork = []

        if totalItems > self.runningWork:
            msg = "Asked to dequeue more work then is running!\n"
            msg += "Failing"
            logging.error(msg)
            raise ProcessPoolException(msg)

        while totalItems > 0:
            try:
                output = self.sink.recv()
                decode = self.jsonHandler.decode(output)
                if isinstance(decode, dict) and decode.get('type', None) == 'ERROR':
                    # Then we had some kind of error
                    msg = decode.get('msg', 'Unknown Error in ProcessPool')
                    logging.error("Received Error Message from ProcessPool Slave")
                    logging.error(msg)
                    self.close()
                    raise ProcessPoolException(msg)
                completedWork.append(decode)
                self.runningWork -= 1
                totalItems -= 1
            except Exception as ex:
                msg = "Exception while getting slave outputin ProcessPool.\n"
                msg += str(ex)
                logging.error(msg)
                break

        return completedWork

    def restart(self):
        """
        _restart_

        Delete everything and restart all pools
        """

        self.close()
        self.createSlaves()
        return


def setupLogging(componentDir):
    """
    _setupLogging_

    Setup logging for the slave process.  Each slave process will have its own
    log file.
    """
    logFile = "%s/ComponentLog.%s" % (componentDir, os.getpid())

    logHandler = RotatingFileHandler(logFile, "a", 1000000000, 3)
    logFormatter = logging.Formatter("%(asctime)s:%(levelname)s:%(module)s:%(message)s")
    logHandler.setFormatter(logFormatter)
    logging.getLogger().addHandler(logHandler)
    logging.getLogger().setLevel(logging.INFO)
    # This is left in as a reminder for debugging purposes
    # SQLDEBUG turns your log files into horrible messes
    # logging.getLogger().setLevel(logging.SQLDEBUG)

    myThread = threading.currentThread()
    myThread.logger = logging.getLogger()
    return


def setupDB(config, wmInit):
    """
    _setupDB_

    Create the database connections.
    """
    socket = getattr(config.CoreDatabase, 'socket', None)
    connectUrl = config.CoreDatabase.connectUrl
    dialect = config.CoreDatabase.dialect

    wmInit.setDatabaseConnection(dbConfig=connectUrl,
                                 dialect=dialect,
                                 socketLoc=socket)
    return


if __name__ == "__main__":
    """
    __main__

    Entry point for the slave process.  The slave's classname will be passed in
    on the command line.  The database connection parameters as well as the
    name of the directory that the log files will be stored in will be passed
    in through stdin as a JSON object.

    Input variables:
    className, input port, output port, path to pickled config, component dir, namespace
    """

    # Get variables passed in
    slaveClassName = sys.argv[1]
    inPort = sys.argv[2]
    outPort = sys.argv[3]
    configPath = sys.argv[4]
    componentDir = sys.argv[5]
    namespace = sys.argv[6]

    # Set up logging
    setupLogging(componentDir)

    # Build ZMQ link
    context = zmq.Context()
    receiver = context.socket(zmq.PULL)
    receiver.connect("tcp://localhost:%s" % inPort)

    sender = context.socket(zmq.PUSH)
    sender.connect("tcp://localhost:%s" % outPort)

    # Build config
    if not os.path.exists(configPath):
        # We can do nothing -
        logging.error("Something in the way of the config path")
        sys.exit(1)

    with open(configPath, 'rb') as f:
        config = pickle.load(f)


    # Setup DB
    wmInit = WMInit()
    setupDB(config, wmInit)

    # Create JSON handler
    jsonHandler = JSONRequests()

    wmFactory = WMFactory(name="slaveFactory", namespace=namespace)
    slaveClass = wmFactory.loadObject(classname=slaveClassName, args=config)

    logging.info("Have slave class")

    while (True):
        encodedInput = receiver.recv()

        try:
            input = jsonHandler.decode(encodedInput)
        except Exception as ex:
            logging.error("Error decoding: %s" % str(ex))
            break

        if input == "STOP":
            break

        try:
            logging.error(input)
            output = slaveClass(input)
        except Exception as ex:
            crashMessage = "Slave process crashed with exception: " + str(ex)
            crashMessage += "\nStacktrace:\n"

            stackTrace = traceback.format_tb(sys.exc_info()[2], None)
            for stackFrame in stackTrace:
                crashMessage += stackFrame

            logging.error(crashMessage)
            try:
                output = {'type': 'ERROR', 'msg': crashMessage}
                encodedOutput = jsonHandler.encode(output)
                if PY3:
                    sender.send_string(encodedOutput)
                else:
                    sender.send(encodedOutput)
                logging.error("Sent error message and now breaking")
                break
            except Exception as ex:
                logging.error("Failed to send error message")
                logging.error(str(ex))
                del jsonHandler
                sys.exit(1)

        if output != None:
            if isinstance(output, list):
                for item in output:
                    encodedOutput = jsonHandler.encode(item)
                    if PY3:
                        sender.send_string(encodedOutput)
                    else:
                        sender.send(encodedOutput)
            else:
                encodedOutput = jsonHandler.encode(output)
                if PY3:
                    sender.send_string(encodedOutput)
                else:
                    sender.send(encodedOutput)

    logging.info("Process with PID %s finished" % (os.getpid()))
    del jsonHandler
    sys.exit(0)



from builtins import str, zip, range, object
from future.utils import viewitems, viewvalues, listvalues

import cherrypy
import inspect
import os
import re
import signal
from string import ascii_letters as letters
import time
from collections import namedtuple
from functools import wraps
from threading import Thread, Condition

from cherrypy import engine, expose, request, response, HTTPError, HTTPRedirect, tools
from cherrypy.lib import cpstats

from WMCore.REST.Error import *
from WMCore.REST.Format import *
from WMCore.REST.Validation import validate_no_more_input
from Utils.CPMetrics import promMetrics

from Utils.Utilities import encodeUnicodeToBytes

try:
    from cherrypy.lib import httputil
except:
    from cherrypy.lib import http as httputil

#: List of HTTP methods for which it's possible to register a REST handler.
_METHODS = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE')

#: Regexp for censoring passwords out of SQL statements before logging them.
_RX_CENSOR = re.compile(r"(identified by) \S+", re.I)

#: MIME types which are compressible.
_COMPRESSIBLE = ['text/html', 'text/html; charset=utf-8',
                 'text/plain', 'text/plain; charset=utf-8',
                 'text/css', 'text/css; charset=utf-8',
                 'text/javascript', 'text/javascript; charset=utf-8',
                 'application/json']

#: Type alias for arguments passed to REST validation methods, consisting
#: of `args`, the additional path arguments, and `kwargs`, the query
#: arguments either from the query string or body (but not both).
RESTArgs = namedtuple("RESTArgs", ["args", "kwargs"])


######################################################################
######################################################################
class RESTFrontPage(object):
    """Base class for a trivial front page intended to hand everything
    over to a javascript-based user interface implementation.

    This front-page simply serves static content like HTML pages, CSS,
    JavaScript and images from number of configurable document roots.
    Text content such as CSS and JavaScript can be scrunched together,
    or combo-loaded, from several files. All the content supports the
    standard ETag and last-modified validation for optimal caching.

    The base class assumes the actual application consists of a single
    front-page, which loads JavaScript and other content dynamically,
    and uses HTML5 URL history features to figure out what to do. That
    is, given application mount point <https://cmsweb.cern.ch/app>, all
    links such as <https://cmsweb.cern.ch/app/foo/bar?q=xyz> get mapped
    to the same page, which then figures out what to do at /foo/bar
    relative location or with the query string part.

    There is a special response for ``rest/preamble.js`` static file.
    This will automatically generate a scriptlet of the following form,
    plus any additional content passed in `preamble`::

      var REST_DEBUG = (debug_mode),
          REST_SERVER_ROOT = "(mount)",
          REST_INSTANCES = [{ "id": "...", "title": "...", "rank": N }...];

    REST_DEBUG, ``debug_mode``
      Is set to true/false depending on the value of the constructor
      :ref:`debug_mode` parameter, or if the default None, it's set
      to false if running with minimised assets, i.e. frontpage matches
      ``*-min.html``, true otherwise.

    REST_SERVER_ROOT, ``mount``
      The URL mount point of this object, needed for history init. Taken
      from the constructor argument.

    REST_INSTANCES
      If the constructor is given `instances`, its return value is turned
      into a sorted JSON list of available instances for JavaScript. Each
      database instance should have the dictionary keys "``.title``" and
      "``.order``" which will be used for human visible instance label and
      order of appearance, respectively. The ``id`` is the label to use
      for REST API URL construction: the instance dictionary key. This
      variable will not be emitted at all if :ref:`instances` is None.

    .. rubric:: Attributes

    .. attribute:: _app

       Reference to the application object given to the constructor.

    .. attribute:: _mount

       The mount point given in the constructor.

    .. attribute:: _static

       The roots given to the constructor, with ``rest/preamble.js`` added.

    .. attribute:: _frontpage

       The name of the front page file.

    .. attribute:: _substitutions

       Extra (name, value) substitutions for `_frontpage`. When serving
       the front-page via `_serve()`, each ``@name@`` is replaced by its
       corresponding `value`.

    .. attribute:: _embeddings

       Extra (name, value) file replacements for `_frontpage`. Similar to
       `_substitutions` but each value is a list of file names, and the
       replacement value is the concatenation of all the contents of all
       the files, with leading and trailing white space removed.

    .. attribute:: _preamble

       The ``rest/preamble.js`` computed as described above.

    .. attribute:: _time

       Server start-up time, used as mtime for ``rest/preamble.js``.
    """

    def __init__(self, app, config, mount, frontpage, roots,
                 substitutions=None, embeddings=None,
                 instances=None, preamble=None, debug_mode=None):
        """.. rubric:: Constructor

        :arg app:                Reference to the :class:`~.RESTMain` application.
        :arg config:             :class:`~.WMCore.ConfigSection` for me.
        :arg str mount:          URL tree mount point for this object.
        :arg str frontpage:      Name of the front-page file, which must exist in
                                 one of the `roots`. If `debug_mode` is None and
                                 the name matches ``*-min.html``, then debug mode
                                 is set to False, True otherwise.
        :arg dict roots:         Dictionary of roots for serving static files.
                                 Each key defines the label and path root for URLs,
                                 and the value should have keys "``root``" for the
                                 path to start looking up files, and "``rx``" for
                                 the regular expression to define valid file names.
                                 **All the root paths must end in a trailing slash.**
        :arg dict substitutions: Extra (name, value) substitutions for `frontpage`.
        :arg dict embeddings:    Extra (name, value) file replacements for `frontpage`.
        :arg callable instances: Callable which returns database instances, often
                                 ``lambda: return self._app.views["data"]._db``
        :arg str preamble:       Optional string for additional content for the
                                 pseudo-file ``rest/preamble.js``.
        :arg bool debug_mode:    Specifies how to set REST_DEBUG, see above."""

        # Verify all roots do end in a slash.
        for origin, info in viewitems(roots):
            if not re.match(r"^[-a-z0-9]+$", origin):
                raise ValueError("invalid root label")
            if not info["root"].endswith("/"):
                raise ValueError("%s 'root' must end in a slash" % origin)

        # Add preamble pseudo-root.
        roots["rest"] = {"root": None, "rx": re.compile(r"^preamble(?:-min)?\.js$")}

        # Save various things.
        self._start = time.time()
        self._app = app
        self._mount = mount
        self._frontpage = frontpage
        self._static = roots
        self._substitutions = substitutions or {}
        self._embeddings = embeddings or {}
        if debug_mode is None:
            debug_mode = not frontpage.endswith("-min.html")

        # Delay preamble setup until server start-up so that we don't try to
        # dereference instances() until after it's been finished constructing.
        engine.subscribe("start", lambda: self._init(debug_mode, instances, preamble), 0)

    def _init(self, debug_mode, instances, preamble):
        """Delayed preamble initialisation after server is fully configured."""
        self._preamble = ("var REST_DEBUG = %s" % ((debug_mode and "true") or "false"))
        self._preamble += (", REST_SERVER_ROOT = '%s'" % self._mount)

        if instances:
            instances = [dict(id=k, title=v[".title"], order=v[".order"])
                         for k, v in viewitems(instances())]
            instances.sort(key=lambda x: x["order"])
            self._preamble += (", REST_INSTANCES = %s" % json.dumps(instances))

        self._preamble += ";\n%s" % (preamble or "")

    def _serve(self, items):
        """Serve static assets.

        Serve one or more files. If there is just one file, it can be text or
        an image. If there are several files, they are smashed together as a
        combo load operation. In that case it's assumed the files are compatible,
        for example all JavaScript or all CSS.

        All normal response headers are set correctly, including Content-Type,
        Last-Modified, Cache-Control and ETag. Automatically handles caching
        related request headers If-Match, If-None-Match, If-Modified-Since,
        If-Unmodified-Since and responds appropriately. The caller should use
        CherryPy gzip tool to handle compression-related headers appropriately.

        In general files are passed through unmodified. The only exception is
        that HTML files will have @MOUNT@ string replaced with the mount point,
        the @NAME@ substitutions from the constructor are replaced by value,
        and @NAME@ embeddings are replaced by file contents.

        :arg list(str) items: One or more file names to serve.
        :returns: File contents combined as a single string."""
        mtime = 0
        result = ""
        ctype = ""

        if not items:
            raise HTTPError(404, "No such file")

        for item in items:
            # There must be at least one slash in the file name.
            if item.find("/") < 0:
                cherrypy.log("ERROR: directory required for front-page '%s'" % item)
                raise HTTPError(404, "No such file")

            # Split the name to the origin part - the name we look up in roots,
            # and the remaining path part for the rest of the name under that
            # root. For example 'yui/yui/yui-min.js' means we'll look up the
            # path 'yui/yui-min.js' under the 'yui' root.
            origin, path = item.split("/", 1)
            if origin not in self._static:
                cherrypy.log("ERROR: front-page '%s' origin '%s' not in any static root"
                             % (item, origin))
                raise HTTPError(404, "No such file")

            # Look up the description and match path name against validation rx.
            desc = self._static[origin]
            if not desc["rx"].match(path):
                cherrypy.log("ERROR: front-page '%s' not matched by rx '%s' for '%s'"
                             % (item, desc["rx"].pattern, origin))
                raise HTTPError(404, "No such file")

            # If this is not the pseudo-preamble, make sure the requested file
            # exists, and if it does, read it and remember its mtime. For the
            # pseudo preamble use the precomputed string and server start time.
            if origin != "rest":
                fpath = desc["root"] + path
                if not os.access(fpath, os.R_OK):
                    cherrypy.log("ERROR: front-page '%s' file does not exist" % item)
                    raise HTTPError(404, "No such file")
                try:
                    mtime = max(mtime, os.stat(fpath).st_mtime)
                    data = open(fpath).read()
                except:
                    cherrypy.log("ERROR: front-page '%s' failed to retrieve file" % item)
                    raise HTTPError(404, "No such file")
            elif self._preamble:
                mtime = max(mtime, self._start)
                data = self._preamble
            else:
                cherrypy.log("ERROR: front-page '%s' no preamble for 'rest'?" % item)
                raise HTTPError(404, "No such file")

            # Concatenate contents and set content type based on name suffix.
            ctypemap = {"js": "text/javascript; charset=utf-8",
                        "css": "text/css; charset=utf-8",
                        "html": "text/html; charset=utf-8"}
            suffix = path.rsplit(".", 1)[-1]
            if suffix in ctypemap:
                if not ctype:
                    ctype = ctypemap[suffix]
                elif ctype != ctypemap[suffix]:
                    ctype = "text/plain"
                if suffix == "html":
                    for var, value in viewitems(self._substitutions):
                        data = data.replace("@" + var + "@", value)
                    for var, files in viewitems(self._embeddings):
                        value = ""
                        for fpath in files:
                            if not os.access(fpath, os.R_OK):
                                cherrypy.log("ERROR: embedded '%s' file '%s' does not exist"
                                             % (var, fpath))
                                raise HTTPError(404, "No such file")
                            try:
                                mtime = max(mtime, os.stat(fpath).st_mtime)
                                value += open(fpath).read().strip()
                            except:
                                cherrypy.log("ERROR: embedded '%s' file '%s' failed to"
                                             " retrieve file" % (var, fpath))
                                raise HTTPError(404, "No such file")
                        data = data.replace("@" + var + "@", value)
                    data = data.replace("@MOUNT@", self._mount)
                if result:
                    result += "\n"
                result += data
                if not result.endswith("\n"):
                    result += "\n"
            elif suffix == "gif":
                ctype = "image/gif"
                result = data
            elif suffix == "png":
                ctype = "image/png"
                result = data
            else:
                raise HTTPError(404, "Unexpected file type")

        # Build final response + headers.
        response.headers['Content-Type'] = ctype
        response.headers['Last-Modified'] = httputil.HTTPDate(mtime)
        response.headers['Cache-Control'] = "public, max-age=%d" % 86400
        response.headers['ETag'] = '"%s"' % hashlib.sha1(encodeUnicodeToBytes(result)).hexdigest()
        cherrypy.lib.cptools.validate_since()
        cherrypy.lib.cptools.validate_etags()
        return result

    @expose
    @tools.gzip(compress_level=9, mime_types=_COMPRESSIBLE)
    def static(self, *args, **kwargs):
        """Serve static assets.

        Assumes a query string in the format used by YUI combo loader, with one
        or more file names separated by ampersands (&). Each name must be a plain
        file name, to be found in one of the roots given to the constructor.

        Path arguments must be empty, or consist of a single 'yui' string, for
        use as the YUI combo loader. In that case all file names are prefixed
        with 'yui/' to make them compatible with the standard combo loading.

        Serves assets as documented in :meth:`_serve`."""
        # The code was originally designed to server YUI content.
        # Modified to support any content by joining args into single path
        # on web page it can be /path/static/js/file.js or /path/static/css/file.css
        if len(args) > 1 or (args and args[0] != "yui"):
            return self._serve(['/'.join(args)])
        # Path arguments must be empty, or consist of a single 'yui' string,
        paths = request.query_string.split("&")
        if not paths:
            raise HTTPError(404, "No such file")
        if args:
            paths = [args[0] + "/" + p for p in paths]
        return self._serve(paths)

    @expose
    def feedback(self, *args, **kwargs):
        """Receive browser problem feedback. Doesn't actually do anything, just
        returns an empty string response."""
        return ""

    @expose
    @tools.gzip(compress_level=9, mime_types=_COMPRESSIBLE)
    def default(self, *args, **kwargs):
        """Generate the front page, as documented in :meth:`_serve`. The
        JavaScript will actually work out what to do with the rest of the
        URL arguments; they are not used here."""
        return self._serve([self._frontpage])

    @expose
    def stats(self):
        """
        Return CherryPy stats dict about underlying service activities
        """
        return cpstats.StatsPage().data()

    @expose
    def metrics(self):
        """
        Return CherryPy stats following the prometheus metrics structure
        """
        metrics = promMetrics(cpstats.StatsPage().data(), self.app.appname)
        return encodeUnicodeToBytes(metrics)


######################################################################
######################################################################
class MiniRESTApi(object):
    """Minimal base class for REST services.

    .. rubric:: Overview

    This is the main base class for the CherryPy-based REST API interfaces.
    It provides a minimal interface for registering API methods to be called
    via REST HTTP calls. Normally implementations should derive from the
    :class:`~.RESTApi` higher level abstraction.

    Instances of this class maintain a table of handlers associated to HTTP
    method (GET, HEAD, POST, PUT, or DELETE) and an API word. The given API
    word may support any number of different HTTP methods by associating
    them to the different, or even same, callables. Each such API may be
    associated with arbitrary additional arguments, typically options to
    adapt behaviour for the API as described below.

    Each API handler must be associated to a validation method which will
    be given the remaining URL path and query arguments. The validator
    must *verify* the incoming arguments and *move* safe input to output
    arguments, typically using :mod:`~.RESTValidation` utilities. If any
    arguments remain in input after validation, the call is refused. This
    implies the only way to get *any input at all* to the REST API is to
    introduce a validator code, which must actively process every piece
    of input to be used later in the actual method.

    When invoked by CherryPy, the class looks up the applicable handler by
    HTTP request method and the first URL path argument, runs validation,
    invokes the method, and formats the output, which should be some python
    data structure which can be rendered into JSON or XML, or some custom
    format (e.g. raw bytes for an image). For sequences and sequence-like
    results, the output is streamed out, object by object, as soon as the
    handler returns; the entire response is never built up in memory except
    as described below for compression and entity tag generation. As the
    assumption is that method will generate some pythonic data structure,
    the caller is required to specify via "Accept" header the desired
    output format.

    .. rubric:: Validation

    There is some general validation before invoking the API method. The
    HTTP method is first checked against allowed methods listed above, and
    those for which there is a registered API handler. The API is checked
    to be known, and that a handler has been registered for the combination
    of the two, and that method can produce output in the format requested
    by in the "Accept" header. The derived class can run :meth:`_precall`
    validation hook before the API name is popped off URL path arguments,
    in case it wants to pull other arguments such as database instance
    label from the URL first.

    The API-specific validation function is called with argument list
    *(apiobj, method, apiname, args, safeargs).* The *apiobj* is the metadata
    object for the API entry created by :meth:`_addAPI`, *method* the HTTP
    method, *apiname* the API name. The remaining two arguments are incoming
    and validated safe arguments, as described above for validation; both
    are instances of :class:`RESTArgs`.

    .. rubric:: Method call

    The API handler is invoked with arguments and keyword arguments saved
    into *safeargs* by the validation method. The method returns a value,
    which must be either a python string, a sequence or a generator.

    .. rubric:: Output formatting

    Once the REST API has returned a value, the class formats the output.
    For GET and HEAD requests, expire headers are set as specified by the
    API object. After that the format handler is invoked, whose output is
    compressed if allowed by the API and the HTTP request, and entity tag
    headers are processed and generated. Although the API may return data
    in any form for which there is a formatter (cf. :class:`~.JSONFormat`,
    :class:`~.XMLFormat` and :class:`~.RawFormat`) that can be matched to
    the request "Accept" header, the general assumption is APIs return
    "rows" of "objects" which are output with "Transfer-Encoding: chunked"
    in streaming format.

    It is supported and recommended that APIs verify the operation will
    very likely succeed, then return a python *generator* for the result,
    for example a database cursor for query results. This allows large
    responses to be streamed out as they are produced, without ever
    buffering the entire response as a string in memory.

    The output from the response formatter is sent to a compressor if the
    request headers include suitable "Accept-Encoding" header and the API
    object hasn't declined compression. Normally all output is compressed
    with ZLIB level 9 provided the client supports ``deflate`` encoding.
    It is recommended to keep compression enabled except if the output is
    known to be incompressible, e.g. images or compressed data files. The
    added CPU use is usually well worth the network communication savings.

    .. rubric:: Entity tags and response caching

    Before discussing caching, we note the following makes the implicit
    assumption that GET and HEAD requests are idempotent and repeatable
    withut ill effects. It is expected only PUT, POST and DELETE have
    side effects.

    All GET and HEAD responses will get an entity tag, or ETag, header. If
    the API does not generate an ETag, a hash digest is computed over the
    serialised response as it's written out, and included in the trailer
    headers of the response.

    If the API can compute an ETag in a better way than hashing the output,
    or can reply to If-Match / If-None-Match queries without computing the
    result, it should absolutely do so. Since most servers generate dynamic
    content such as database query results, it's hard to generate a useful
    ETag, and the scheme used here, a digest over the content, allows many
    clients like browsers to cache the responses. The server still has to
    re-execute the query and reformat the result to recompute the ETag, but
    avoids having to send matching responses back, and more importantly, the
    client learns its cached copy remains valid and can optimise accordingly.

    As a usability and performance optimisation, small GET/HEAD responses up
    to :attr:`etag_limit` bytes are buffered entirely, the ETag is computed,
    any If-Match, If-None-Match request headers are processed, the ETag is
    added to the response headers, and the entire body is output as a single
    string without the regular chunked streaming. This effectively allows
    web browser clients to cache small responses even if the API method is
    unable to compute a reliable ETag for it. This covers virtually all
    responses one would access in a browser for most servers in practice.

    The scheme also reduces network traffic since the body doesn't need to be
    output for successful cache validation queries, and when body is output,
    TCP tends to get fed more efficiently with larger output buffers. On the
    other hand the scheme degrades gracefully for large replies while allowing
    smarter clients, such as those using curl library to talk to the server,
    to get more out of the API, both performance and functionality wise.

    For large responses the computed digest ETag is still added as a trailer
    header, after all the output has been emitted. These will not be useful
    for almost any client. For one, many clients ignore trailers anyway. For
    another there is no way to perform If-Match / If-None-Match validation.

    The default digest algorithm is :class:`~.SHA1ETag`, configurable via
    :attr:`etagger`. The tagging and compression work also if the API uses
    CherryPy's ``serve_file()`` to produce the response.

    For GET responses to be cached by clients the response must include a
    both a "Cache-Control" header and an ETag (or "Last-Modified" time). If
    :attr:`default_expires` and :attr:`default_expires_opts` settings allow
    it, GET responses will include a "Cache-Control: max-age=n" header.
    These can be tuned per API with ``cherrypy.tools.expires(secs=n)``, or
    ``expires`` and ``expires_opts`` :func:`restcall` keyword arguments.

    .. rubric:: Notes

    .. note:: Only GET and HEAD requests are allowed to have a query string.
       Other methods (POST, PUT, DELETE) may only have parameters specified
       in the request body. This is protection against XSRF attacks and in
       general attempts to engineer someone to submit a malicious HTML form.

    .. note:: The ETag generated by the default hash digest is computed from
       the formatted stream, before compression. It is therefore independent
       of "Accept-Encoding" request headers. However the ETag value is stable
       only if the formatter produces stable output given the same input. In
       particular any output which includes python dictionaries may vary over
       calls because of changes in dictionary key iteration order.

    .. warning:: The response generator returned by the API may fail after it
       has started to generate output. If the response is large enough not
       to fit in the ETag buffer, there is really no convenient way to stop
       producing output and indicate an error to the client since the HTTP
       headers have already been sent out with a good chunk of normal output.

       If the API-returned generator throws, this implementation closes the
       generated output by adding any required JSON, XML trailers and sets
       X-REST-Status and normal error headers in the trailer headers. The
       clients that understand trailer headers, such as curl-based ones, can
       use the trailer to discover the output is incomplete. Clients which
       ignore trailers, such as web browsers and python's urllib and httplib,
       will not know the output is truncated. Hence derived classes should be
       conservative about using risky constructs in the generator phase.

       Specifically any method with side effects should return a generator
       response only *after* all side effects have already taken place. For
       example when updating a database, the client should commit first,
       then return a generator. Because of how generators work in python,
       the latter step must be done in a separate function from the one that
       invokes commit!

       None of this really matters for responses which fit into the ETag
       buffer after compression. That is all but the largest responses in
       most servers, and the above seems a reasonable trade-off between
       not having to buffer huge responses in memory and usability.

    .. rubric:: Attributes

    .. attribute:: app

       Reference to the :class:`~.RESTMain` object from constructor.

    .. attribute:: config

       Reference to the :class:`WMCore.Configuraation` section for this API
       mount point as passed in to the constructor by :class:`~.RESTMain`.

    .. attribute:: methods

       A dictionary of registered API methods. The keys are HTTP methods,
       the values are another dictionary level with API name as a key, and
       an API object as a value. Do not modify this table directly, use the
       :meth:`_addAPI` method instead.

    .. attribute:: formats

       Possible output formats matched against "Accept" headers. This is a
       list of *(mime-type, formatter)* tuples. This is the general list of
       formats supported by this API entry point, with the intention that
       all or at least most API methods support these formats. Individual
       API methods can override the format list using ``formats`` keyword
       argument to :func:`restcall`, for example to specify list of image
       formats for one method which retrieves image files
       (cf. :class:`~.RawFormat`).

    .. attribute:: etag_limit

       An integer threshold for number of bytes to buffer internally for
       calculation of ETag header. See the description above for the full
       details on the scheme. The API object can override this limit with
       the ``etag_limit`` keyword argument to :func:`restcall`. The default
       is 8 MB.

    .. attribute:: compression

       A list of accepted compression mechanisms to be matched against the
       "Accept-Encoding" HTTP request header. Currently supported values are
       ``deflate`` and ``identity``. Using ``identity`` or emptying the list
       disables compression. The default is ``['deflate']``. Change this only
       for API mount points which are known to generate incompressible output,
       using ``compression`` keyword argument to :func:`restcall`.

    .. attribute:: compression_level

       Integer 0-9, the default ZLIB compression level for ``deflate`` encoding.
       The default is the maximum 9; for most servers the increased CPU use is
       usually well worth the reduction in network transmission costs. Setting
       the level to zero disables compression. The API can override this value
       with ``compression_level`` keyword argument to :func:`restcall`.

    .. attribute:: compression_chunk

       Integer, the approximate amount of input, in bytes, to consume to form
       compressed output chunks. The preferred value is the ZLIB compression
       horizon, 64kB. Up to this many bytes of output from the stream formatter
       are fed to the compressor, after which the compressor is flushed and the
       compressed output is emitted as HTTP transmission level chunk. Hence the
       receiving side can process chunks much like if the data wasn't compressed:
       the only difference is each HTTP chunk will contain an integral number of
       chunks instead of just one. (Of course in most cases the ETag buffering
       scheme will consume the output and emit it as a single unchunked part.)
       The API can override this value with ``compression_chunk`` keyword
       argument to :func:`restcall`.

    .. attribute:: default_expires

       Number, default expire time for GET / HEAD responses in seconds. The
       API object can override this value with ``expires`` keyword argument
       to :func:`restcall`. The default is one hour, or 3600 seconds.

    .. attribute:: default_expires_opts

       A sequence of strings, additional default options for "Cache-Control"
       response headers for responses with non-zero expire time limit. The
       strings are joined comma-separated to the "max-age=n" item. The API
       object can override this value with ``expires_opts`` keyword argument
       to :func:`restcall`. The default is an empty list.

    .. rubric:: Constructor arguments

    :arg app: The main application :class:`~.RESTMain` object.
    :arg config: The :class:`~.WMCore.ConfigSection` for this object.
    :arg str mount: The CherryPy URL tree mount point for this object.
    """

    def __init__(self, app, config, mount):
        self.app = app
        self.config = config
        self.etag_limit = 8 * 1024 * 1024
        self.compression_level = 9
        self.compression_chunk = 64 * 1024
        self.compression = ['deflate', 'gzip']
        self.formats = [('application/json', JSONFormat()),
                        ('application/xml', XMLFormat(self.app.appname))]
        self.methods = {}
        self.default_expires = 3600
        self.default_expires_opts = []

    def _addAPI(self, method, api, callable, args, validation, **kwargs):
        """Add an API method.

        Use this method to register handlers for a method/api combination. This
        creates an internal "API object" which internally represents API target.
        The API object is dictionary and will be passed to validation functions.

        :arg str method: The HTTP method name: GET, HEAD, PUT, POST, or DELETE.
        :arg str api: The API label, to be matched with first URL path argument.
          The label may not contain slashes.
        :arg callable callable: The handler; see class documentation for signature.
        :arg list args: List of valid positional and keyword argument names.
          These will be copied to the API object which will be passed to the
          validation methods. Normally you'd get these with inspect.getargspec().
        :arg callable validation: The validator; see class documentation for the
          signature and behavioural requirements. If `args` is non-empty,
          `validation` is mandatory; if `args` is empty, `callable` does not
          receive any input. The validator must copy validated safe input to
          actual arguments to `callable`.
        :arg dict kwargs: Additional key-value pairs to set in the API object.

        :returns: Nothing."""

        if method not in _METHODS:
            raise UnsupportedMethod()

        if method not in self.methods:
            self.methods[method] = {}

        if api in self.methods[method]:
            raise ObjectAlreadyExists()

        if not isinstance(args, list):
            raise TypeError("args is required to be a list")

        if not isinstance(validation, list):
            raise TypeError("validation is required to be a list")

        if args and not validation:
            raise ValueError("non-empty validation required for api taking arguments")

        apiobj = {"args": args, "validation": validation, "call": callable}
        apiobj.update(**kwargs)
        self.methods[method][api] = apiobj

    @expose
    def stats(self):
        """
        Return CherryPy stats dict about underlying service activities
        """
        return cpstats.StatsPage().data()

    @expose
    def metrics(self):
        """
        Return CherryPy stats following the prometheus metrics structure
        """
        metrics = promMetrics(cpstats.StatsPage().data(), self.app.appname)
        return encodeUnicodeToBytes(metrics)

    @expose
    def default(self, *args, **kwargs):
        """The HTTP request handler.

        This just wraps `args` and `kwargs` into a :class:`RESTArgs` and invokes
        :meth:`_call` enclosed in a try/except which filters all exceptions but
        :class:`HTTPRedirect` via :func:`~.report_rest_error`.

        In other words the main function of this wrapper is to ensure run-time
        errors are properly logged and translated to meaningful response status
        and headers, including a trace identifier for this particular error. It
        also sets X-REST-Time response header to the total time spent within
        this request.

        This method is declared "``{response.stream: True}``" to CherryPy.

        :returns: See :meth:`_call`."""

        try:
            return self._call(RESTArgs(list(args), kwargs))
        except HTTPRedirect:
            raise
        except Exception as e:
            report_rest_error(e, format_exc(), True)
        finally:
            if getattr(request, 'start_time', None):
                response.headers["X-REST-Time"] = "%.3f us" % \
                                                  (1e6 * (time.time() - request.start_time))

    default._cp_config = {'response.stream': True}

    def _call(self, param):
        """The real HTTP request handler.

        :arg RESTArgs param: Input path and query arguments.

        :returns: Normally a generator over the input, but in some cases a
          plain string, as described in the class-level documentation. The
          API handler response, often a generator itself, is usually wrapped
          in couple of layers of additional generators which format and
          compress the output, and handle entity tags generation and matching."""

        # Make sure the request method is something we actually support.
        if request.method not in self.methods:
            response.headers['Allow'] = " ".join(sorted(self.methods.keys()))
            raise UnsupportedMethod() from None

        # If this isn't a GET/HEAD request, prevent use of query string to
        # avoid cross-site request attacks and evil silent form submissions.
        # We'd prefer perl cgi behaviour where query string and body args remain
        # separate, but that's not how cherrypy works - it mixes everything
        # into one big happy kwargs.
        if (request.method != 'GET' and request.method != 'HEAD') \
                and request.query_string:
            response.headers['Allow'] = 'GET HEAD'
            raise MethodWithoutQueryString()

        # Give derived class a chance to look at arguments.
        self._precall(param)

        # Make sure caller identified the API to call and it is available for
        # the request method.
        if len(param.args) == 0:
            raise APINotSpecified()
        api = param.args.pop(0)
        if api not in self.methods[request.method]:
            methods = " ".join(sorted([m for m, d in viewitems(self.methods) if api in d]))
            response.headers['Allow'] = methods
            if not methods:
                msg = 'Api "%s" not found. This method supports these Apis: %s' % (api, list(self.methods[request.method]))
                raise APINotSupported(msg)
            else:
                msg = 'Api "%s" only supported in method(s): "%s"' % (api, methods)
                raise APIMethodMismatch(msg)
        apiobj = self.methods[request.method][api]

        # Check what format the caller requested. At least one is required; HTTP
        # spec says no "Accept" header means accept anything, but that is too
        # error prone for a REST data interface as that establishes a default we
        # cannot then change later. So require the client identifies a format.
        # Browsers will accept at least */*; so can clients who don't care.
        # Note that accept() will raise HTTPError(406) if no match is found.
        # Available formats are either specified by REST method, or self.formats.
        try:
            if not request.headers.elements('Accept'):
                raise NotAcceptable('Accept header required')
            formats = apiobj.get('formats', self.formats)
            format = cherrypy.lib.cptools.accept([f[0] for f in formats])
            fmthandler = [f[1] for f in formats if f[0] == format][0]
        except HTTPError:
            format_names = ', '.join(f[0] for f in formats)
            raise NotAcceptable('Available types: %s' % format_names)

        # Validate arguments. May convert arguments too, e.g. str->int.
        safe = RESTArgs([], {})
        for v in apiobj['validation']:
            v(apiobj, request.method, api, param, safe)
        validate_no_more_input(param)

        # Invoke the method.
        obj = apiobj['call'](*safe.args, **safe.kwargs)

        # Add Vary: Accept header.
        vary_by('Accept')

        # Set expires header if applicable. Note that POST/PUT/DELETE are not
        # cacheable to begin with according to HTTP/1.1 specification. We must
        # do this before actually streaming out the response below in case the
        # ETag matching decides the previous response remains valid.
        if request.method == 'GET' or request.method == 'HEAD':
            expires = self.default_expires
            cpcfg = getattr(apiobj['call'], '_cp_config', None)
            if cpcfg and 'tools.expires.on' in cpcfg:
                expires = cpcfg.get('tools.expires.secs', expires)
            expires = apiobj.get('expires', expires)
            if 'Cache-Control' in response.headers:
                pass
            elif expires < 0:
                response.headers['Pragma'] = 'no-cache'
                response.headers['Expires'] = 'Sun, 19 Nov 1978 05:00:00 GMT'
                response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate,' \
                                                    ' post-check=0, pre-check=0'
            elif expires != None:
                expires_opts = apiobj.get('expires_opts', self.default_expires_opts)
                expires_opts = (expires_opts and ', '.join([''] + expires_opts)) or ''
                response.headers['Cache-Control'] = 'max-age=%d%s' % (expires, expires_opts)

        # Format the response.
        response.headers['X-REST-Status'] = 100
        response.headers['Content-Type'] = format
        etagger = apiobj.get('etagger', None) or SHA1ETag()
        reply = stream_compress(fmthandler(obj, etagger),
                                apiobj.get('compression', self.compression),
                                apiobj.get('compression_level', self.compression_level),
                                apiobj.get('compression_chunk', self.compression_chunk))
        return stream_maybe_etag(apiobj.get('etag_limit', self.etag_limit), etagger, reply)

    def _precall(self, param):
        """Point for derived classes to hook into prior to peeking at URL.

        Derived classes receive the incoming path and keyword arguments in
        `param`, and can peek or modify those as appropriate. In particular
        if they want to pop something off the path part of the URL before
        the base class takes the API name off it, this is the time to do it.

        The base class implementation does nothing at all.

        :arg RESTArgs param: Input path and query arguments.
        :returns: Nothing."""
        pass


######################################################################
######################################################################
class RESTApi(MiniRESTApi):
    """REST service using :class:`~.RESTEntity` for API implementation.

    This is base class for a REST API implemented in terms of *entities*
    which support GET/PUT/POST/DELETE methods, instead of registering
    individual API methods. The general principle is to perform entity
    modelling, with express relationships, a little like databases are
    designed, then expose those entities as REST API points. This is
    usually very different from exposing just a collection of RPC APIs.

    Normally each entity represents a collection of objects, and HTTP
    methods translate to customary meanings over the collection:
    GET = query collection, PUT = insert new, POST = update existing,
    DELETE = remove item. Often the entity is written to be natively
    array-oriented so that a single PUT can insert a large number of
    objects, or single DELETE can remove many objects, for example.
    Using optional URL path arguments the entity can support operations
    on sub-items as well as collection-wide ones.

    Entities should be registered with :meth:`_add`. Where an entity API
    cannot reasonably represent the needs, raw RPC-like methods can still
    be added using :meth:`_addAPI`."""

    def _addEntities(self, entities, entry, wrapper=None):
        """Private interface for adding APIs for entities.

        :arg dict entities: See :meth:`_add` documentation.
        :arg callable entry: The pre-entry callback hook. See :meth:`_enter`.
        :arg callable wrapper: Optional wrapper to modify the method handler.
        :returns: Nothing."""
        for label, entity in viewitems(entities):
            for method in _METHODS:
                handler = getattr(entity, method.lower(), None)
                if not handler and method == 'HEAD':
                    handler = getattr(entity, 'get', None)
                if handler and getattr(handler, 'rest.exposed', False):
                    rest_args = getattr(handler, 'rest.args')
                    rest_params = getattr(handler, 'rest.params')
                    if wrapper: handler = wrapper(handler)
                    self._addAPI(method, label, handler, rest_args,
                                 [entity.validate, entry],
                                 entity=entity, **rest_params)

    def _add(self, entities):
        """Add entities.

        Adds a collection of entities and their labels. Each entity should be
        derived from :class:`~.RESTEntity` and can define a method per HTTP
        method it supports: :meth:`get` to support GET, :meth:`post` for POST,
        and so on. In addition it must provide a :meth:`validate` method
        compatible with the :class:`~.MiniRESTApi` validator convention; it
        will be called for all HTTP methods. If the entity defines :meth:`get`
        but no :meth:`head` method, :meth:`get` is automatically registered as
        the HEAD request handler as well.

        The signature of the HTTP method handlers depends on arguments copied
        to `safe` by the validation function. If :meth:`validate` copies one
        safe argument `foo` to `safe`, then :meth:`get` will be invoked with
        exactly one argument, `foo`. There is no point in specifying default
        values for HTTP method handlers as any defaults must be processed in
        :meth:`validate`. In any case it's better to keep all argument defaults
        in one place, in the :meth:`validate` method.

        :arg dict entities: A dictionary of labels with :class:`~.RESTEntity`
          value. The label is the API name and should be a bare word, with
          no slashes. It's recommended to use dash (-) as a word separator.
          The value is the entity to add, as described above.
        :returns: Nothing."""
        self._addEntities(entities, self._enter)

    def _enter(self, apiobj, method, api, param, safe):
        """Callback to be invoked after validation completes but before
        the actual API method is invoked.

        This sets the ``rest_generate_data`` and ``rest_generate_preamble``
        attributes on ``cherrypy.request``, for use later in output formatting.

        If the :func:`restcall` for the API method was called with ``generate``
        property, it's set as ``request.rest_generate_data`` for use in output
        formatting to label the name of the result object from this call. The
        default value is ``"result"``.

        The ``request.rest_generate_preamble`` is used to add a description to
        the response preamble. By default there is no description, but if the
        :func:`restcall` call for the API method used ``columns`` property, its
        value is set as ``columns`` key to the preamble. This would be used if
        the output is inherently row oriented, and the preamble gives the column
        titles and subsequent output the column values. This is the recommended
        REST API output format, especially for large output, as it is easy to
        read in clients but much more compact than emitting column labels for
        every row, e.g. by using dictionaries for each row.

        The API may modify ``request.rest_generate_preamble`` if it wishes to
        insert something to the description part.

        :arg dict apiobj: API object from :class:`~.MiniRESTApi`.
        :arg str method: HTTP method.
        :arg str api: Name of the API being called.
        :arg RESTArgs param: Incoming arguments.
        :arg RESTArgs safe: Output safe arguments.
        :returns: Nothing."""
        request.rest_generate_data = apiobj.get("generate", None)
        request.rest_generate_preamble = {}
        cols = apiobj.get("columns", None)
        if cols:
            request.rest_generate_preamble["columns"] = cols


######################################################################
######################################################################
class DBConnectionPool(Thread):
    """Asynchronous and robust database connection pool.

    .. rubric:: The pool

    This class provides a database connection pool that is thread safe
    and recovers as gracefully as possible from server side and network
    outages. Thread safety means that multiple threads may check out and
    return connections from or to the pool conncurrently. All connection
    management operations are internally transferred to a separate thread
    in order to guarantee client web server threads never block even if
    the database layer blocks or hangs in connection-related calls.

    In other words the effective purpose of this class is to guarantee
    web servers autonomously and gracefully enter a degraded state when
    the underlying database or network goes out, responding with "503
    database unavailable" but continuing to serve requests otherwise.
    Once database returns to operation the servers normally recover on
    their own without having to be manually restarted. While the main
    intent is to avoid requiring intervention on a production service,
    as a side effect this class makes database-based web services more
    usable on mobile devices with unstable network connections.

    The primary pooling purpose is to cache connections for databases
    for which connections are expensive, in particular Oracle. Instead
    of creating a new connection for each use, much of the time the pool
    returns a compatible idle connection which the application was done
    using. Connections unused for more than a pool-configured timeout are
    automatically closed.

    Between uses the connections are reset by rolling back any pending
    operations, and tested for validity by probing the server for
    "liveness" response. Connections failing tests are automatically
    disposed; when returning conections to the pool clients may indicate
    they've had trouble with it, and the pool will reclaim the connection
    instead of queuing it for reuse. In general it is hard to tell which
    errors involved connection-related issues, so it is safer to flag all
    errors on returning the connection. It is however better to filter
    out frequent benign errors such as integrity violations to avoid
    excessive connectin churn.

    Other database connection pool implementations exist, including for
    instance a session pool in the Oracle API (cx_Oracle's SessionPool).
    The reason this class implements pooling and not just validation is
    to side-step blocking or thread unsafe behaviour in the others. All
    connection management operations are passed to a worker thread. If
    the database connection layer stalls, HTTP server threads requesting
    connections will back off gracefully, reporting to their own clients
    that the database is currently unavailable. On the other hand no
    thread is able to create a connection to database if one connection
    stalls, but usually that would happen anyway with at least Oracle.

    The hand-over to the worker thread of course adds overhead, but the
    increased robustness and gracefulness in face of problems in practice
    outweighs the cost by far, and is in any case cheaper than creating
    a new connection each time. The level of overhead can be tuned by
    adjusting condition variable contention (cf. `num_signals`).

    The connections returned to clients are neither garbage collected
    nor is there a ceiling on a maximum number of connections returned.
    The client needs to be careful to `put()` as many connections as it
    received from `get()` to avoid leaking connections.

    .. rubric:: Pool specifications

    The database specification handed to the constructor should be a
    dictionary with the members:

    ``type``
      Reference to the DB API module, aka connection type.

    ``schema``
      String, name of the database schema the connection references.
      Sets connection ``current_schema`` attribute.

    ``clientid``
      String, identifies the client to session monitor, normally this
      should be `service-label@fully.qualified.domain`. Sets connection
      ``client_identifier`` attribute.

    ``liveness``
      String, SQL to execute to verify the connection remain usable,
      normally "``select sysdate from dual``" or alike. The statement
      should require a fresh response from the server on each execution
      so avoid "``select 1 from dual``" style cacheable queries.

    ``user``
      String, login user name.

    ``password``
      String, login password for ``user``.

    ``dsn``
      String, login data source name, usually the TNS entry name.

    ``timeout``
      Integer or float, number of seconds to retain unused idle
      connections before closing them. Note that this only applies to
      connections handed back to `put()`; connections given by `get()`
      but never returned to the pool are not considered idle, not even
      if the client loses the handle reference.

    ``stmtcache``
      Optional integer, overrides the default statement cache size 50.

    ``trace``
      Optional boolean flag, if set enables tracing of database activity
      for this pool, including connection ops, SQL executed, commits and
      rollbacks, and pool activity on the handles. If True, connections
      are assigned random unique labels which are used in activity log
      messages. Recycled idle connections also get a new label, but the
      connection log message will include the old label which allows
      previous logs on that connection to be found. Not to be confused
      with database server side logging; see ``session-sql`` for that.

    ``auth-role``
      Optional, (NAME, PASSWORD) string sequence. If set, connections
      acquire the database role *NAME* before use by executing the SQL
      "``set role none``", "``set role NAME identified by PASSWORD``"
      on each `get()` of the connection. In other words, if the role is
      removed or the password is changed, the client will automatically
      shed the role and fail with an error, closing the connection in
      the process, despite connection caching.

    ``session-sql``
      Optional sequence of SQL statement strings. These are executed on
      each connection `get()`. Use with session trace statements such as
      "``alter session set sql_trace = true``",
      "``alter session set events '10046 trace name context forever,
      level 12'``". It's not recommended to make any database changes.

    .. rubric:: Connection handles

    The `get()` method returns a database connection handle, a dict with
    the following members. The exact same dict needs to be returned to
    `put()` -- not a copy of it.

    ``type``
      Reference to DB API module.

    ``pool``
      Reference to this pool object.

    ``connection``
      Reference to the actual DB API connection object.

    ``trace``
      Always present, but may be either boolean False, or a non-empty
      string with the trace message prefix to use for all operations
      concerning this connection.

    .. rubric:: Attributes

    .. attribute:: connection_wait_time

       Number, the maximum time in seconds to wait for a connection to
       complete after which the client will be told the database server
       is unavailable.  This should be large enough to avoid triggering
       unnecessary database unavailability errors in sporadic delays in
       production use, but low enough to bounce clients off when the
       database server is having difficulty making progress.

       In particular client HTTP threads will be tied up this long if
       the DB server is completely hanging: completing TCP connections
       but not the full database handshake, or if TCP connection itself
       is experiencing significant delays. Hence it's important to keep
       this value low enough for the web server not to get dogged down
       or fail with time-out errors itself.

    .. attribute:: wakeup_period

       Number, maximum time in seconds to wait in the worker thread main
       loop before checking any timed out connections. The server does
       automatically adjust the wake-up time depending on work needed,
       so there usually isn't any need to change this value. The value
       should not be decreased very low to avoid an idle server from
       waking up too often.

    .. attribute:: num_signals

       Number of condition variables to use for signalling connection
       completion. The pool creates this many condition variables and
       randomly picks one to signal connection completion between the
       worker and calling threads. Increase this if there is a high
       degree of thread contention on concurrent threads waiting for
       connection completion. The default should be fine for all but
       highest connection reuse rates.

    .. attribute:: max_tries

       The maximum number of times to attempt creating a connection
       before giving up. If a connection fails tests, it is discarded
       and another attempt is made with another connection, either an
       idle one or an entirely new connection if no idle ones remain.
       This variable sets the limit on how many times to try before
       giving up. This should be high enough a value to consume any
       cached bad connections rapidly enough after network or database
       failure. Hence the pool will reclaim any bad connections at the
       maximum rate of `get()` calls times `max_tries` per
       `connection_wait_time`.

       Do not set this value to a very high value if there is a real
       possibility of major operational flukes leading to connection
       storms or account lock-downs, such as using partially incorrect
       credentials or applications with invalid/non-debugged SQL which
       cause connection to be black-listed and recycled. In other words,
       only change this parameter for applications which have undergone
       significant testing in a production environment, with clear data
       evidence the default value is not leading to sufficiently fast
       recovery after connections have started to go sour.

    .. attribute:: dbspec

       Private, the database specification given to the constructor.

    .. attribute:: id

       Private, the id given to the constructor for trace logging.

    .. attribute:: sigready

       Private, `num_signals` long list of condition variables for
       signalling connection attempt result.

    .. attribute:: sigqueue

       Private, condition variable for signalling changes to `queue`.

    .. attribute:: queue

       Private, list of pending requests to the worker thread, access
       to which is protected by `sigqueue`. Connection release requests
       go to the front of the list, connection create requests at the
       end. The worker thread takes the first request in queue, then
       executes the action with `sigqueue` released so new requests can
       be added while the worker is talking to the database.

    .. attribute:: inuse

       Private, list of connections actually handed out by `get()`. Note
       that if the client has already given up on the `get()` request by
       the time the connection is finally established, the connection is
       automatically discarded and not put no this list. This list may be
       accessed only in the worker thread as no locks are available to
       protect the access; `logstatus()` method provides the means to log
       the queue state safely in the worker thread.

    .. attribute:: idle

       Private, list of idle connections, each of which has ``expires``
       element to specify the absolute time when it will expire. The
       worker thread schedules to wake up within five seconds after the
       next earliest expire time, or in `wakeup_period` otherwise, and
       of course whenever new requests are added to `queue`. This list
       may be accessed only in the work thread as no locks are available
       to protect the access; `logstatus()` method provides the means to
       log the queue state safely in the worker thread.

    .. rubric:: Constructor

    The constructor automatically attaches this object to the cherrypy
    engine start/stop messages so the background worker thread starts or
    quits, respectively. The pool does not attempt to connect to the
    database on construction, only on the first call to `get()`, so it's
    safe to create the pool even if network or database are unavailable.

    :arg dict dbspec: Connection specification as described above.
    :arg str id: Identifier used to label trace connection messages for
                 this pool, such as the full class name of the owner."""

    connection_wait_time = 8
    wakeup_period = 60
    num_signals = 4
    max_tries = 5

    def __init__(self, id, dbspec):
        Thread.__init__(self, name=self.__class__.__name__)
        self.sigready = [Condition() for _ in range(0, self.num_signals)]
        self.sigqueue = Condition()
        self.queue = []
        self.idle = []
        self.inuse = []
        if type in dbspec and dbspec['type'].__name__ == 'MySQLdb':
            dbspec['dsn'] = dbspec['db']
        self.dbspec = dbspec
        self.id = id
        engine.subscribe("start", self.start, 100)
        engine.subscribe("stop", self.stop, 100)

    def logstatus(self):
        """Pass a request to the worker thread to log the queue status.

        It's recommended that the owner hook this method to a signal such
        as SIGUSR2 so it's possible to get the process dump its database
        connections status, especially the number of `inuse` connections,
        from outside the process.

        The request is inserted in the front of current list of pending
        requests, but do note the request isn't executed directly. If the
        worker thread is currently blocked in a database or network call,
        log output is only generated when the worker resumes control.

        :returns: Nothing."""
        self.sigqueue.acquire()
        self.queue.insert(0, (self._status, None))
        self.sigqueue.notifyAll()
        self.sigqueue.release()

    def stop(self):
        """Tell the pool to stop processing requests and to exit from the
        worker thread.

        The request is inserted in the front of current list of pending
        requests. The worker thread will react to it as soon as it's done
        processing any currently ongoing database or network call. If the
        database API layer is completely wedged, that might be never, in
        which case the application should arrange for other means to end,
        either by using a suicide alarm -- for example by calling
        signal.alarm() but not setting SIGALRM handler -- or externally
        by arranging SIGKILL to be delivered.

        Since this request is inserted in the front of pending requests,
        existing connections, whether idle or in use, will not be closed
        or even rolled back. It's assumed the database server will clean
        up the connections once the process exits.

        The constructor automatically hooks the cherrypy engine 'stop'
        message to call this method.

        :returns: Nothing."""
        self.sigqueue.acquire()
        self.queue.insert(0, (None, None))
        self.sigqueue.notifyAll()
        self.sigqueue.release()

    def get(self, id, module):
        """Get a new connection from the pool, identified to server side and
        the session monitor as to be used for action `id` by `module`.

        This retrieves the next available idle connection from the pool, or
        creates a new connection if none are available. Before handing back
        the connection, it's been tested to be actually live and usable.
        If the database connection specification included a role attribute
        or session statements, they will have been respectively set and
        executed.

        The connection request is appended to the current queue of requests.
        If the worker thread does not respond in `connection_wait_time`, the
        method gives up and indicates the database is not available. When
        that happens, the worker thread will still attempt to complete the
        connection, but will then discard it.

        :arg str id:     Identification string for this connection request.
                         This will set the ``clientinfo`` and ``action``
                         attributes on the connection for database session
                         monitoring tools to visualise and possibly remote
                         debugging of connection use.

        :arg str module: Module using this connection, typically the fully
                         qualified python class name. This will set the
                         ``module`` attribute on the connection object for
                         display in database session monitoring tools.

        :returns: A `(HANDLE, ERROR)` tuple. If a connection was successfully
                  made, `HANDLE` will contain a dict with connection data as
                  described in the class documentation and `ERROR` is `None`.
                  If no connection was made at all, returns `(None, None)`.
                  Returns `(None, (ERROBJ, TRACEBACK))` if there was an error
                  making the connection that wasn't resolved in `max_tries`
                  attempts; `ERROBJ` is the last exception thrown, `TRACEBACK`
                  the stack trace returned by `format_exc()` for it."""

        sigready = random.choice(self.sigready)
        arg = {"error": None, "handle": None, "signal": sigready,
               "abandoned": False, "id": id, "module": module}

        self.sigqueue.acquire()
        self.queue.append((self._connect, arg))
        self.sigqueue.notifyAll()
        self.sigqueue.release()

        sigready.acquire()
        now = time.time()
        until = now + self.connection_wait_time
        while True:
            dbh = arg["handle"]
            err = arg["error"]
            if dbh or err or now >= until:
                arg["abandoned"] = True
                break
            sigready.wait(until - now)
            now = time.time()
        sigready.release()
        return dbh, err

    def put(self, dbh, bad=False):
        """Add a database handle `dbh` back to the pool.

        Normally `bad` would be False and the connection is added back to
        the pool as an idle connection, and will be reused for a subsequent
        connection.

        Any pending operations on connections are automatically cancalled
        and rolled back before queuing them into the idle pool. These will
        be executed asynchronously in the database connection worker thread,
        not in the caller's thread. However note that if the connection
        became unusable, attempting to roll it back may block the worker.
        That is normally fine as attempts to create new connections will
        start to fail with timeout, leading to "database unavailable" errors.

        If the client has had problems with the connection, it should most
        likely set `bad` to True, so the connection will be closed and
        discarded. It's safe to reuse connections after benign errors such
        as basic integrity violations. However there are a very large class
        of obscure errors which actually mean the connection handle has
        become unusable, so it's generally safer to flag the handle invalid
        on error -- with the caveat that errors should be rare to avoid
        excessive connection churn.

        :arg dict dbh: A database handle previously returned by `get()`. It
                       must be the exact same dict object, not a copy.
        :arg bool bad: If True, `dbh` is likely bad, so please try close it
                       instead of queuing it for reuse.
        :returns: Nothing."""

        self.sigqueue.acquire()
        self.queue.insert(0, ((bad and self._disconnect) or self._release, dbh))
        self.sigqueue.notifyAll()
        self.sigqueue.release()

    def run(self):
        """Run the connection management thread."""

        # Run forever, pulling work from "queue". Round wake-ups scheduled
        # from timeouts to five-second quantum to maximise the amount of
        # work done per round of clean-up and reducing wake-ups.
        next = self.wakeup_period
        while True:
            # Whatever reason we woke up, even if sporadically, process any
            # pending requests first.
            self.sigqueue.acquire()
            self.sigqueue.wait(max(next, 5))
            while self.queue:
                # Take next action and execute it. "None" means quit. Release
                # the queue lock while executing actions so callers can add
                # new requests, e.g. release connections while we work here.
                # The actions are not allowed to throw any exceptions.
                action, arg = self.queue.pop(0)
                self.sigqueue.release()
                if action:
                    action(arg)
                else:
                    return
                self.sigqueue.acquire()
            self.sigqueue.release()

            # Check idle connections for timeout expiration. Calculate the
            # next wake-up as the earliest expire time, but note that it
            # gets rounded to minimum five seconds above to scheduling a
            # separate wake-up for every handle. Note that we may modify
            # 'idle' while traversing it, so need to clone it first.
            now = time.time()
            next = self.wakeup_period
            for old in self.idle[:]:
                if old["expires"] <= now:
                    self.idle.remove(old)
                    self._disconnect(old)
                else:
                    next = min(next, old["expires"] - now)

    def _status(self, *args):
        """Action handler to dump the queue status."""
        cherrypy.log("DATABASE CONNECTIONS: %s@%s %s: timeout=%d inuse=%d idle=%d"
                     % (self.dbspec["user"], self.dbspec["dsn"], self.id,
                        self.dbspec["timeout"], len(self.inuse), len(self.idle)))

    def _error(self, title, rest, err, where):
        """Internal helper to generate error message somewhat similar to
        :func:`~.report_rest_error`.

        :arg str title: All-capitals error message title part.
        :arg str rest: Possibly non-empty trailing error message part.
        :arg Exception err: Exception object reference.
        :arg str where: Traceback for `err` as returned by :ref:`format_exc()`."""
        errid = "%032x" % random.randrange(1 << 128)
        cherrypy.log("DATABASE THREAD %s ERROR %s@%s %s %s.%s %s%s (%s)"
                     % (title, self.dbspec["user"], self.dbspec["dsn"], self.id,
                        getattr(err, "__module__", "__builtins__"),
                        err.__class__.__name__, errid, rest, str(err).rstrip()))
        for line in where.rstrip().split("\n"):
            cherrypy.log("  " + line)

    def _connect(self, req):
        """Action handler to fulfill a connection request."""
        s = self.dbspec
        dbh, err = None, None

        # If tracing, issue log line that identifies this connection series.
        trace = s["trace"] and ("RESTSQL:" + "".join(random.sample(string.ascii_letters, 12)))
        trace and cherrypy.log("%s ENTER %s@%s %s (%s) inuse=%d idle=%d" %
                               (trace, s["user"], s["dsn"], self.id, req["id"],
                                len(self.inuse), len(self.idle)))

        # Attempt to connect max_tries times.
        for _ in range(0, self.max_tries):
            try:
                # Take next idle connection, or make a new one if none exist.
                # Then test and prepare that connection, linking it in trace
                # output to any previous uses of the same object.
                dbh = (self.idle and self.idle.pop()) or self._new(s, trace)
                assert dbh["pool"] == self
                assert dbh["connection"]
                prevtrace = dbh["trace"]
                dbh["trace"] = trace
                self._test(s, prevtrace, trace, req, dbh)

                # The connection is ok. Kill expire limit and return this one.
                if "expires" in dbh:
                    del dbh["expires"]
                break
            except Exception as e:
                # The connection didn't work, report and remember this exception.
                # Note that for every exception reported for the server itself
                # we may report up to max_tries exceptions for it first. That's
                # a little verbose, but it's more useful to have all the errors.
                err = (e, format_exc())
                self._error("CONNECT", "", *err)
                dbh and self._disconnect(dbh)
                dbh = None

        # Return the result, and see if the caller abandoned this attempt.
        req["signal"].acquire()
        req["error"] = err
        req["handle"] = dbh
        abandoned = req["abandoned"]
        req["signal"].notifyAll()
        req["signal"].release()

        # If the caller is known to get our response, record the connection
        # into 'inuse' list. Otherwise discard any connection we made.
        if not abandoned and dbh:
            self.inuse.append(dbh)
        elif abandoned and dbh:
            cherrypy.log("DATABASE THREAD CONNECTION ABANDONED %s@%s %s"
                         % (self.dbspec["user"], self.dbspec["dsn"], self.id))
            self._disconnect(dbh)

    def _new(self, s, trace):
        """Helper function to create a new connection with `trace` identifier."""
        trace and cherrypy.log("%s instantiating a new connection" % trace)
        ret = {"pool": self, "trace": trace, "type": s["type"]}
        if s['type'].__name__ == 'MySQLdb':
            ret.update({"connection": s["type"].connect(s['host'], s["user"], s["password"], s["db"], int(s["port"]))})
        else:
            ret.update({"connection": s["type"].connect(s["user"], s["password"], s["dsn"], threaded=True)})

        return ret

    def _test(self, s, prevtrace, trace, req, dbh):
        """Helper function to prepare and test an existing connection object."""
        # Set statement cache. Default is 50 statments but spec can override.
        c = dbh["connection"]
        c.stmtcachesize = s.get("stmtcache", 50)

        # Emit log message to identify this connection object. If it was
        # previously used for something else, log that too for detailed
        # debugging involving problems with connection reuse.
        if s['type'].__name__ == 'MySQLdb':
            client_version = s["type"].get_client_info()
            version = ".".join(str(x) for x in s["type"].version_info)
        else:
            client_version = ".".join(str(x) for x in s["type"].clientversion())
            version = c.version
        prevtrace = ((prevtrace and prevtrace != trace and
                      " (previously %s)" % prevtrace.split(":")[1]) or "")
        trace and cherrypy.log("%s%s connected, client: %s, server: %s, stmtcache: %d"
                               % (trace, prevtrace, client_version,
                                  version, c.stmtcachesize))

        # Set the target schema and identification attributes on this one.
        c.current_schema = s["schema"]
        c.client_identifier = s["clientid"]
        c.clientinfo = req["id"]
        c.module = req["module"]
        c.action = req["id"][:32]

        # Ping the server. This will detect some but not all dead connections.
        trace and cherrypy.log("%s ping" % trace)
        c.ping()

        # At least server responded, now try executing harmless SQL but one
        # that requires server to actually respond. This detects remaining
        # bad connections.
        trace and cherrypy.log("%s check [%s]" % (trace, s["liveness"]))
        c.cursor().execute(s["liveness"])

        # If the pool requests authentication role, set it now. First reset
        # any roles we've acquired before, then attempt to re-acquire the
        # role. Hence if the role is deleted or its password is changed by
        # application admins, we'll shed any existing privileges and close
        # the connection right here. This ensures connection pooling cannot
        # be used to extend role privileges forever.
        if "auth-role" in s:
            trace and cherrypy.log("%s set role none")
            c.cursor().execute("set role none")
            trace and cherrypy.log("%s set role %s" % (trace, s["auth-role"][0]))
            c.cursor().execute("set role %s identified by %s" % s["auth-role"])

        # Now execute session statements, e.g. tracing event requests.
        if "session-sql" in s:
            for sql in s["session-sql"]:
                trace and cherrypy.log("%s session-sql [%s]" % (trace, sql))
                c.cursor().execute(sql)

        # OK, connection's all good.
        trace and cherrypy.log("%s connection established" % trace)

    def _release(self, dbh):
        """Action handler to release a connection back to the pool."""
        try:
            # Check the handle didn't get corrupted.
            assert dbh["pool"] == self
            assert dbh["connection"]
            assert dbh in self.inuse
            assert dbh not in self.idle
            assert "expires" not in dbh

            # Remove from 'inuse' list first in case the rest throws/hangs.
            s = self.dbspec
            trace = dbh["trace"]
            self.inuse.remove(dbh)

            # Roll back any started transactions. Note that we don't want to
            # call cancel() on the connection here as it will most likely just
            # degenerate into useless "ORA-25408: can not safely replay call".
            trace and cherrypy.log("%s release with rollback" % trace)
            dbh["connection"].rollback()

            # Record expire time and put to end of 'idle' list; _connect()
            # takes idle connections from the back of the list, so we tend
            # to reuse most recently used connections first, and to prune
            # the number of connections in use to the minimum.
            dbh["expires"] = time.time() + s["timeout"]
            self.idle.append(dbh)
            trace and cherrypy.log("%s RELEASED %s@%s timeout=%d inuse=%d idle=%d"
                                   % (trace, s["user"], s["dsn"], s["timeout"],
                                      len(self.inuse), len(self.idle)))
        except Exception as e:
            # Something went wrong, nuke the connection from orbit.
            self._error("RELEASE", " failed to release connection", e, format_exc())

            try:
                self.inuse.remove(dbh)
            except ValueError:
                pass

            try:
                self.idle.remove(dbh)
            except ValueError:
                pass

            self._disconnect(dbh)

    def _disconnect(self, dbh):
        """Action handler to discard the connection entirely."""
        try:
            # Assert internal consistency invariants; the handle may be
            # marked for use in case it's discarded with put(..., True).
            assert dbh not in self.idle

            try:
                self.inuse.remove(dbh)
            except ValueError:
                pass

            # Close the connection.
            s = self.dbspec
            trace = dbh["trace"]
            trace and cherrypy.log("%s disconnecting" % trace)
            dbh["connection"].close()

            # Remove references to connection object as much as possible.
            del dbh["connection"]
            dbh["connection"] = None

            # Note trace that this is now gone.
            trace and cherrypy.log("%s DISCONNECTED %s@%s timeout=%d inuse=%d idle=%d"
                                   % (trace, s["user"], s["dsn"], s["timeout"],
                                      len(self.inuse), len(self.idle)))
        except Exception as e:
            self._error("DISCONNECT", " (ignored)", e, format_exc())


######################################################################
######################################################################
class DatabaseRESTApi(RESTApi):
    """A :class:`~.RESTApi` whose entities represent database contents.

    This class wraps API calls into an environment which automatically sets
    up and tears down a database connection around the call, and translates
    all common database errors such as unique constraint violations into an
    appropriate and meaningful :class:`~.RESTError`-derived error.

    This class is fundamentally instance-aware. The :meth:`_precall` hook is
    used to pop a required instance argument off the URL, before the API name,
    so URLs look like ``/app/prod/entity``. The instance name from the URL is
    used to select the database instance on which the API operates, and the
    HTTP method is used to select suitable connection pool, for example reader
    account for GET operations and a writer account for PUT/POST/DELETE ones.

    Normally the entities do not use database connection directly, but use
    the convenience functions in this class. These methods perform many
    utility tasks needed for high quality implementation, plus make it very
    convenient to write readable REST entity implementations which conform
    to the preferred API conventions. For example the :meth:`query` method
    not only executes a query, but it runs a SQL statement cleaner, does all
    the SQL execution trace logging, keeps track of last statement executed
    with its binds (to be logged with error output in case an exception is
    raised later on), fills ``rest_generate_preamble["columns"]`` from the
    SQL columns titles, and returns a generator over the result, optionally
    filtering the results by a regular expression match criteria.

    The API configuration must include ``db`` value, whose value should be
    the qualified name of the python object for the multi-instance database
    authentication specification. The object so imported should yield a
    dictionary whose keys are the supported instance names. The values should
    be dictionaries whose keys are HTTP methods and values are dictionaries
    specifying database contact details for that use; "*" can be used as a
    key for "all other methods". For example the following would define two
    instances, "prod" and "dev", with "prod" using a reader account for GETs
    and a writer account for other methods, and "dev" with a single account::

      DB = { "prod": { "GET": { "user": "foo_reader", ... },
                       "*":   { "user": "foo_writer", ... } },
             "dev":  { "*":   { "user": "foo", ... } } }

    The class automatically uses a thread-safe, resilient connection pool
    for the database connections (cf. :class:`~.DBConnectionPool`). New
    database connections are created as needed on demand. If the database is
    unavailable, the server will gracefully degrade to reporting HTTP "503
    Service Unavailable" status code. Database availability is not required
    on server start-up, connections are first attempted on the first HTTP
    request which requires a connection. In fact connections are prepared
    only *after* request validation has completed, mainly as a security
    measure to prevent unsafe parameter validation code from mixing up with
    database query execution. At the end of API method execution connection
    handles are automatically rolled back, so API should explicitly commit
    if it wants any changes made to last. Sending the server SIGUSR2 signal
    will log connection usage statistics and pool timeouts.

    .. rubric:: Attributes

    .. attribute:: _db

       The database configuration imported from ``self.config.db``. Each
       connection specification has dictionary key ``"pool"`` added,
       pointing to the :class:`~.DBConnectionPool` managing that pool.

    .. rubric:: Constructor

    Constructor arguments are the same as for the base class.
    """
    _ALL_POOLS = []

    def __init__(self, app, config, mount):
        RESTApi.__init__(self, app, config, mount)
        signal.signal(signal.SIGUSR2, self._logconnections)
        modname, item = config.db.rsplit(".", 1)
        module = __import__(modname, globals(), locals(), [item])
        self._db = getattr(module, item)
        myid = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        for spec in viewvalues(self._db):
            for db in viewvalues(spec):
                if isinstance(db, dict):
                    db["pool"] = DBConnectionPool(myid, db)
                    DatabaseRESTApi._ALL_POOLS.append(db["pool"])

    @staticmethod
    def _logconnections(*args):
        """SIGUSR2 signal handler to log status of all pools."""
        list([p.logstatus() for p in DatabaseRESTApi._ALL_POOLS])

    def _add(self, entities):
        """Add entities.

        See base class documentation for the semantics. This wraps the API
        method with :meth:`_wrap` to handle database related errors, and
        inserts :meth:`_dbenter` as the last validator in chain to wrap
        the method in database connection management. :meth:`_dbenter`
        will install :meth:`_dbexit` as a request clean-up callback to
        handle connection disposal if database connection succeeds."""
        self._addEntities(entities, self._dbenter, self._wrap)

    def _wrap(self, handler):
        """Internal helper function to wrap calls to `handler` inside a
        database exception filter. Any exceptions raised will be passed to
        :meth:`_dberror`."""

        @wraps(handler)
        def dbapi_wrapper(*xargs, **xkwargs):
            try:
                return handler(*xargs, **xkwargs)
            except Exception as e:
                self._dberror(e, format_exc(), False)

        return dbapi_wrapper

    def _dberror(self, errobj, trace, inconnect):
        """Internal helper routine to process exceptions while inside
        database code.

        This method does necessary housekeeping to roll back and dispose any
        connection already made, remember enough of the context for error
        reporting, and convert the exception into a relevant REST error type
        (cf. :class:`~.RESTError`). For example constraint violations turn
        into :class:`~.ObjectAlreadyExists` or :class:`~.MissingObject`.

        Errors not understood become :class:`~.DatabaseExecutionError`, with
        the SQL and bind values of last executed statement reported into the
        server logs (but not to the client). Note that common programming errors,
        :class:`~.KeyError`, :class:`~.ValueError` or :class:`~.NameError`
        raised in API implementation, become :class:`~.DatabaseExecutionError`
        too. This will make little difference to the client, and will still
        include all the relevant useful information in the server logs, but
        of course they are not database-related as such.

        :arg Exception errobj: The exception object.
        :arg str trace: Associated traceback as returned by :func:`format_exc`.
        :arg bool inconnect: Whether exception occurred while trying to connect.
        :return: Doesn't, raises an exception."""

        # Grab last sql executed and whatever binds were used so far,
        # the database type object, and null out the rest so that
        # post-request and any nested error handling will ignore it.
        db = request.db
        type = db["type"]
        instance = db["instance"]
        sql = (db["last_sql"],) + db["last_bind"]

        # If this wasn't connection failure, just release the connection.
        # If that fails, force drop the connection. We ignore errors from
        # this since we are attempting to report another earlier error.
        db["handle"] and db["pool"].put(db["handle"], True)

        # Set the db backend
        DB_BACKEND = db['type'].__name__

        del request.db
        del db

        # Raise an error of appropriate type.
        errinfo = {"errobj": errobj, "trace": trace}
        dberrinfo = {"errobj": errobj, "trace": trace,
                     "lastsql": sql, "instance": instance}
        if inconnect:
            raise DatabaseUnavailable(**dberrinfo)
        elif isinstance(errobj, type.IntegrityError):
            errorcode = errobj[0] if DB_BACKEND == 'MySQLdb' else errobj.args[0].code
            if errorcode in {'cx_Oracle': (1, 2292), 'MySQLdb': (1062,)}[DB_BACKEND]:
                # ORA-00001: unique constraint (x) violated
                # ORA-02292: integrity constraint (x) violated - child record found
                # MySQL: 1062, Duplicate entry 'x' for key 'y'
                # MySQL: Both unique and integrity constraint falls into 1062 error
                raise ObjectAlreadyExists(**errinfo)
            elif errorcode in {'cx_Oracle': (1400, 2290), 'MySQLdb': (1048,)}[DB_BACKEND]:
                # ORA-01400: cannot insert null into (x)
                # ORA-02290: check constraint (x) violated
                # MySQL: 1048, Column (x) cannot be null
                # There are no check constraint in MySQL. Oracle 2290 equivalent does not exist
                raise InvalidParameter(**errinfo)
            elif errorcode == {'cx_Oracle': 2291, 'MySQLdb': 1452}[DB_BACKEND]:
                # ORA-02291: integrity constraint (x) violated - parent key not found
                # MySQL: 1452, Cannot add or update a child row: a foreign key constraint fails
                raise MissingObject(**errinfo)
            else:
                raise DatabaseExecutionError(**dberrinfo)
        elif isinstance(errobj, type.OperationalError):
            raise DatabaseUnavailable(**dberrinfo)
        elif isinstance(errobj, type.InterfaceError):
            raise DatabaseConnectionError(**dberrinfo)
        elif isinstance(errobj, (HTTPRedirect, RESTError)):
            raise
        else:
            raise DatabaseExecutionError(**dberrinfo)

    def _precall(self, param):
        """Pop instance from the URL path arguments before :meth:`_call` takes
        entity name from it.

        This orders URL to have entities inside an instance, which is logical.
        Checks the instance name is present and one of the known databases as
        registered to the constructor into :attr:`_db`, and matches the HTTP
        method to the right database connection pool.

        If the instance argument is missing, unsafe, or unknown, raises an
        :class:`~.NoSuchInstance` exception. If there is no connection pool
        for that combination of HTTP method and instance, raises an
        :class:`~.DatabaseUnavailable` exception.

        If successful, remembers the database choice in ``cherrypy.request.db``
        but does not yet attempt connection to it. This is a dictionary with
        the following fields.

        ``instance``
          String, the instance name.

        ``type``
          Reference to the DB API module providing database connection.

        ``pool``
          Reference to the :class:`~.DBConnectionPool` providing connections.

        ``handle``
          Reference to the database connection handle from the pool. Initially
          set to `None`.

        ``last_sql``
          String, the last SQL statement executed on this connection. Initially
          set to `None`, filled in by the statement execution utility function
          :meth:`prepare` (which :meth:`execute` and :meth:`executemany` call).
          Used for error reporting in :meth:`_dberror`.

        ``last_bind``
          A tuple *(values, keywords)* of the last bind values used on this
          connection. Initially set to *(None, None)* and reset back to that
          value by each call to :meth:`prepare`, filled in by the statement
          execution utility functions :meth:`executemany` and :meth:`execute`.
          Use for error reporting in :meth:`_dberror`. Note that if the data
          is security sensitive, or extremely voluminous with thousands of
          values, this might cause some concern.

        :arg RESTArgs param: Incoming URL path and query arguments. If a valid
          instance name is found, pops the first path item off ``param.args``.
        :returns: Nothing."""

        # Check we were given an instance argument and it's known.
        if not param.args or param.args[0] not in self._db:
            raise NoSuchInstance()
        if not re.match(r"^[a-z]+$", param.args[0]):
            raise NoSuchInstance("Invalid instance name")

        instance = param.args.pop(0)

        # Get database object.
        if request.method in self._db[instance]:
            db = self._db[instance][request.method]
        elif request.method == 'HEAD' and 'GET' in self._db[instance]:
            db = self._db[instance]['GET']
        elif '*' in self._db[instance]:
            db = self._db[instance]['*']
        else:
            raise DatabaseUnavailable()

        # Remember database instance choice, but don't do anything about it yet.
        request.db = {"instance": instance, "type": db["type"], "pool": db["pool"],
                      "handle": None, "last_sql": None, "last_bind": (None, None)}

    def _dbenter(self, apiobj, method, api, param, safe):
        """Acquire database connection just before invoking the entity.

        :meth:`_add` arranges this method to be invoked as the last validator
        function, just before the base class calls the actual API method. We
        acquire the database connection here, and arrange the connection to
        be released by :meth:`_dbexit` after the response has been generated.

        It is not possible to release the database connection right after the
        response handler returns, as it is normally a generator which is still
        holding on to a database cursor for streaming out the response. Here we
        register connection release to happen in the CherryPy request clean-up
        phase ``on_end_request``. Clean-up is ensured even if errors occurred.

        Database connections will be identified by the HTTP method, instance
        name and entity name, and with the fully qualified class name of the
        entity handling the request. This allows server activity to be more
        usefully monitored in the database session monitor tools.

        Raises a :class:`~.DatabaseUnavailable` exception if connecting to the
        database fails, either the connection times out or generates errors.
        This exception is translated into "503 Service Unavailable" HTTP status.
        Note that :class:`~.DBConnectionPool` verifies the connections are
        actually usable before handing them back, so connectivity problems are
        normally all caught here, with the server degrading into unavailability.

        A successfully made connection is recorded in ``request.db["handle"]``
        and ``request.rest_generate_data`` is initialised from the ``generate``
        :func:`restcall` parameter, as described in :meth:`.RESTApi._enter`.
        ``request.rest_generate_preamble`` is reset to empty; it will normally
        be filled in by :meth:`execute` and :meth:`executemany` calls.

        :args: As documented for validation functions.
        :returns: Nothing."""

        assert getattr(request, "db", None), "Expected DB args from _precall"
        assert isinstance(request.db, dict), "Expected DB args from _precall"

        # Get a pool connection to the instance.
        request.rest_generate_data = None
        request.rest_generate_preamble = {}
        module = "%s.%s" % (apiobj['entity'].__class__.__module__,
                            apiobj['entity'].__class__.__name__)
        id = "%s %s %s" % (method, request.db["instance"], api)
        dbh, err = request.db["pool"].get(id, module)

        if err:
            self._dberror(err[0], err[1], True)
        elif not dbh:
            del request.db
            raise DatabaseUnavailable()
        else:
            request.db["handle"] = dbh
            request.rest_generate_data = apiobj.get("generate", None)
            request.hooks.attach('on_end_request', self._dbexit, failsafe=True)

    def _dbexit(self):
        """CherryPy request clean-up handler to dispose the database connection.

        Releases the connection back to the :class:`~.DBConnectionPool`. As
        described in :meth:`_dbenter`, call to this function is arranged at
        the end of the request processing inside CherryPy.

        :returns: Nothing."""

        if getattr(request, "db", None) and request.db["handle"]:
            request.db["pool"].put(request.db["handle"], False)

    def sqlformat(self, schema, sql):
        """Database utility function which reformats the SQL statement by
        removing SQL ``--`` comments and leading and trailing white space
        on lines, and converting multi-line statements to a single line.

        The main reason this function is used is that often SQL statements
        are written in code as multi-line quote blocks, and it's awkward
        to include them in database tracing and error logging output. It
        is much more readable to have single line statements with canonical
        spacing, so we put all executed statements into that form before use.

        If `schema` is provided, the statement is munged to insert its value
        as a schema prefix to all tables, sequences, indexes and keys that
        are not already prefixed by a schema name. This is not normally used,
        the preferred mechanism is to set ``current_schema`` attribute on the
        connections as per :class:`~.DBConnectionPool` documentation. For the
        munging to work, it is assumed all table names start with ``t_``, all
        sequence names start with ``seq_``, all index names start with ``ix_``,
        all primary key names start with ``pk_`` and all foreign key names
        start with ``fk_``.

        This method does not remove ``/* ... */`` comments to avoid removing
        any deliberate query hints.

        :arg str sql: SQL statement to clean up.
        :arg str schema: If not `None`, schema prefix to insert.
        :returns: Cleaned up SQL statement string."""

        sql = re.sub(r"--.*", "", sql)
        sql = re.sub(r"^\s+", "", sql)
        sql = re.sub(r"\s+$", "", sql)
        sql = re.sub(r"\n\s+", " ", sql)
        if schema:
            sql = re.sub(r"(?<=\s)((t|seq|ix|pk|fk)_[A-Za-z0-9_]+)(?!\.)",
                         r"%s.\1" % schema, sql)
            sql = re.sub(r"(?<=\s)((from|join)\s+)([A-Za-z0-9_]+)(?=$|\s)",
                         r"\1%s.\3" % schema, sql)
        return sql

    def prepare(self, sql):
        """Prepare a SQL statement.

        This utility cleans up SQL statement `sql` with :meth:`sqlformat`,
        obtains a new cursor from the current database connection, and
        calls the ``prepare()`` on it, then returns the cursor.

        The cleaned up statement is remembered as ``request.db["last_sql"]``
        and if database tracing is enabled, logged for the current connection.
        In log output, sensitive contents like passwords are censored out
        (cf. :obj:`~._RX_CENSOR`). The binds in ``request.db["last_bind"]``
        are reset to (None, None).

        This method can be called only between :meth:`_dbenter` and
        :meth:`_dbexit`.

        :arg str sql: SQL statement.
        :returns: Cursor on which statement was prepared."""

        assert request.db["handle"], "DB connection missing"
        sql = self.sqlformat(None, sql)  # FIXME: schema prefix?
        trace = request.db["handle"]["trace"]
        logsql = re.sub(_RX_CENSOR, r"\1 <censored>", sql)
        request.db["last_bind"] = None, None
        request.db["last_sql"] = logsql
        trace and cherrypy.log("%s prepare [%s]" % (trace, logsql))
        c = request.db["handle"]["connection"].cursor()
        if request.db['type'].__name__ == 'MySQLdb':
            return c
        c.prepare(sql)
        return c

    def execute(self, sql, *binds, **kwbinds):
        """Execute a SQL statement with bind variables.

        This method mirrors the DB API :func:`execute` method on cursors. It
        executes the `sql` statement after invoking :meth:`prepare` on it,
        saving the bind variables into ``request.db["last_bind"]`` and
        logging the binds if tracing is enabled. It returns both the cursor
        and the return value from the underlying DB API call as a tuple.

        The convention for passing binds is the same as for the corresponding
        DB API :func:`execute` method on cursors. You may find :meth:`bindmap`
        useful to convert dict-of-lists keyword arguments of a REST entity to
        the commonly used list-of-dicts to `binds` argument for this method.

        This method can be called only between :meth:`_dbenter` and
        :meth:`_dbexit`.

        :arg str sql: SQL statement string.
        :arg list binds: Positional binds.
        :arg dict kwbinds: Keyword binds.
        :returns: Tuple *(cursor, variables)*, where *cursor* is the cursor
          returned by :meth:`prepare` for the SQL statement, and *variables*
          is the list of variables if `sql` is a query, or None, as returned
          by the corresponding DB API :func:`execute` method."""

        c = self.prepare(sql)
        trace = request.db["handle"]["trace"]
        request.db["last_bind"] = (binds, kwbinds)
        trace and cherrypy.log("%s execute: %s %s" % (trace, binds, kwbinds))
        if request.db['type'].__name__ == 'MySQLdb':
            return c, c.execute(sql, kwbinds)
        return c, c.execute(None, *binds, **kwbinds)

    def executemany(self, sql, *binds, **kwbinds):
        """Execute a SQL statement many times with bind variables.

        This method mirrors the DB DBI :func:`executemany` method on cursors.
        It executes the statement over a sequence of bind values; otherwise
        it is the same as :meth:`execute`.

        :args: See :meth:`execute`.
        :returns: See :meth:`execute`."""

        c = self.prepare(sql)
        trace = request.db["handle"]["trace"]
        request.db["last_bind"] = (binds, kwbinds)
        trace and cherrypy.log("%s executemany: %s %s" % (trace, binds, kwbinds))
        if request.db['type'].__name__ == 'MySQLdb':
            return c, c.executemany(sql, binds[0])
        return c, c.executemany(None, *binds, **kwbinds)

    def query(self, match, select, sql, *binds, **kwbinds):
        """Convenience function to :meth:`execute` a query, set ``"columns"`` in
        the REST response preamble to the column titles of the query result, and
        return a generator over the cursor as a result, possibly filtered.

        The SQL statement is expected to be a SELECT query statement. The `sql`,
        `binds` and `kwbinds` are passed to :meth:`execute`. The returned cursor's
        ``description``, the column titles of the returned rows, are used to set
        ``request.rest_generate_preamble["columns"]`` description preamble. The
        assumption is those column titles are compatible with the argument names
        accepted by REST methods such that "columns" returned here are compatible
        with the argument names accepted by the GET/PUT/POST/DELETE methods.

        If `match` is not `None`, it is assumed to be a regular expression object
        to be passed to :func:`rxfilter` to filter results by column value, and
        `select` should be a callable which returns the value to filter on given
        the row values, typically an :func:`operator.itemgetter`.  If `match` is
        `None`, this function returns the trivial generator :func:`rows`.

        The convention for passing binds is the same as for the corresponding
        DB API :func:`execute` method on cursors. You may find :meth:`bindmap`
        useful to convert dict-of-lists keyword arguments of a REST entity to
        the commonly used list-of-dicts to `binds` argument for this method.

        This method can be called only between :meth:`_dbenter` and
        :meth:`_dbexit`.

        :arg re.RegexObject match: Either `None` or a regular expression for
          filtering the results by a colum value.
        :arg callable select: If `match` is not `None`, a callable which
          retrieves the column value to run `match` against. It must return
          a string value. It would normally be :func:`operator.itemgetter`,
          may need additional code if type conversion is needed, or if the
          caller wants to join values of multiple columns for filtering.
        :arg str sql: SQL statement string.
        :arg list binds: Positional binds.
        :arg dict kwbinds: Keyword binds.
        :returns: A generator over the query results."""

        c, _ = self.execute(sql, *binds, **kwbinds)
        request.rest_generate_preamble["columns"] = \
            [x[0].lower() for x in c.description]
        if match:
            return rxfilter(match, select, c)
        else:
            return rows(c)

    def modify(self, sql, *binds, **kwbinds):
        """Convenience function to :meth:`executemany` a query, and return a
        a result with the number of affected rows, after verifying it matches
        exactly the number of inputs given and committing the transaction.

        The SQL statement is expected to be a modifying query statement: INSERT,
        UPDATE, DELETE, MERGE, and so on. If `binds` is non-empty, it is passed
        as-is to :meth:`executemany` with `kwbinds`, and the expected number of
        rows to be affected is set to *len(binds[0]).* Otherwise it's assumed
        arguments are given as lists by keyword, and they are transposed to a
        list of dictionaries with :meth:`bindmap`; for REST methods this is
        expected to be the most common calling convention since REST API calls
        themselves will receive arguments as lists by keyword argument. The
        expected number of affected rows is set to the length of the list
        returned by :meth:`bindmap`; note that :meth:`bindmap` requires all
        arguments to have the same number of values.

        Since :meth:`executemany` is used, the operation is intrinsically array
        oriented. This property is useful for making REST methods natively
        collection oriented, allowing operations to act efficiently on large
        quantities of input data with very little special coding.

        After executing the statement but before committing, this method runs
        :meth:`rowstatus` to check that the number of rows affected matches
        exactly the number of inputs. :meth:`rowstatus` throws an exception
        if that is not the case, and otherwise returns a trivial result object
        of the form ``[{ "modified": rowcount }]`` meant to be returned to REST
        API callers.

        Once :meth:`rowstatus` is happy with the result, the transaction is
        committed, with appropriate debug output if trace logging is enabled.

        :arg str sql: SQL modify statement.
        :arg list binds: Bind variables by position: list of dictionaries.
        :arg dict kwbinds: Bind variables by keyword: dictionary of lists.
        :result: See :meth:`rowstatus` and description above."""

        if binds:
            c, _ = self.executemany(sql, *binds, **kwbinds)
            expected = len(binds[0])
        else:
            kwbinds = self.bindmap(**kwbinds)
            c, _ = self.executemany(sql, kwbinds, *binds)
            expected = len(kwbinds)
        result = self.rowstatus(c, expected)
        trace = request.db["handle"]["trace"]
        trace and cherrypy.log("%s commit" % trace)
        request.db["handle"]["connection"].commit()
        return result

    def rowstatus(self, c, expected):
        """Verify the last statement executed on cursor `c` touched exactly
        `expected` number of rows.

        The last SQL statement executed on `c` is assumed to have been a
        modifying such as INSERT, DELETE, UPDATE or MERGE. If the number of
        rows affected by the statement is less than `expected`, raises a
        :class:`~.MissingObject` exception. If the number is greater than
        `expected`, raises a :class:`~.TooManyObjects` exception. If the
        number is exactly right, returns a generator over a trivial result
        object of the form ``[{ "modified": rowcount }]`` which is meant to
        be returned as the result from the underlying REST method.

        Normally you would use this method via :meth:`modify`, but it can be
        called directly to verify each step of multi-statement update operation.

        This method exists so that a REST entity can be written easily to have
        a documented invariant, which is likely to always pass, and needs to
        be checked with very low overhead -- aka each modify operation touches
        the exact number of database objects requested, and attempts to act on
        non-existent objects are caught and properly reported to clients. It
        enables trivially one to use ``INSERT INTO ... SELECT ...`` statement
        style for insertions, deletions and updates. If the SELECT part returns
        fewer or more results than expected, an exception will be raised and the
        transaction rolled back even if the schema constraints do not catch the
        problem.

        In other words the check here makes sure that any app logic or database
        schema constraint mistakes turn into hard API errors rather than making
        hash out of the database contents, or lying to clients that an operation
        succeeded when it did not actually fully perform the requested task. It
        eliminates the need to proliferate result checking code through the app
        and to worry about a large class of possible race conditions in clustered
        web services talking to databases.

        This method can be called on any valid cursor object. It's in no way
        dependent on the use or non-use of the other database utility methods.

        :arg c: DB API cursor object.
        :arg int expected: Number of rows expected to be affected.
        :returns: See description above."""

        if c.rowcount < expected:
            raise MissingObject(info="%d vs. %d expected" % (c.rowcount, expected))
        elif c.rowcount > expected:
            raise TooManyObjects(info="%d vs. %d expected" % (c.rowcount, expected))
        return rows([{"modified": c.rowcount}])

    def bindmap(self, **kwargs):
        """Given `kwargs` of equal length list keyword arguments, returns the
        data transposed as list of dictionaries each of which has a value for
        every key from each of the lists.

        This method is convenient for arranging HTTP request keyword array
        parameters for bind arrays suitable for `executemany` call.

        For example the call ``api.bindmap(a = [1, 2], b = [3, 4])`` returns a
        list of dictionaries ``[{ "a": 1, "b": 3 }, { "a": 2, "b": 4 }]``."""

        keys = list(kwargs)
        return [dict(list(zip(keys, vals))) for vals in zip(*listvalues(kwargs))]


######################################################################
######################################################################
class RESTEntity(object):
    """Base class for entities in :class:`~.RESTApi`-based interfaces.

    This class doesn't offer any service other than holding on to the
    arguments given in the constructor. The derived class must implement at
    least a :meth:`validate` method, and some number of HTTP request method
    functions as described in :class:`~.RESTApi` documentation.

    Do note that keyword arguments with defaults are pointless on HTTP method
    handlers: the methods are never called in a fashion that would actually
    use the declaration defaults. Specifically, every argument stored into
    ``safe.kwargs`` by a validator will always be given *some* value. If the
    parameter was declared optional to the validator, it will automatically
    be given `None` value when not present in the request arguments.

    .. rubric:: Attributes

    .. attribute:: app

       Reference to the :class:`~.RESTMain` application.

    .. attribute:: api

       Reference to the :class:`~.RESTApi` owner of this entity.

    .. attribute:: config

       Reference to the :class:`WMCore.ConfigSection` for the
       :class:`~.RESTApi` owner.

    .. attribute:: mount

       The URL mount point of the :class:`~.RESTApi` owner. Does not
       include the name of this entity.
    """

    def __init__(self, app, api, config, mount):
        self.app = app
        self.api = api
        self.config = config
        self.mount = mount


######################################################################
######################################################################
def restcall(func=None, args=None, generate="result", **kwargs):
    """Mark a method for use in REST API calls.

    This is a decorator to mark a callable, such as :class:`~.RESTEntity`
    method, an exposed REST interface. It must be applied on functions and
    methods given to :meth:`.MiniRESTApi._addAPI` and :meth:`.RESTApi._add`.
    It can be used either as bare-word or function-like decorator: both
    ``@restcall def f()`` and ``@restcall(foo=bar) def f()`` forms work.

    The `args` should be the parameter names accepted by the decorated
    function. If `args` is the default None, the possible arguments are
    extracted with :func:`inspect.getargspec`.

    The `generate` argument sets the label used by output formatters to
    surround the output. The default is "result", yielding for example
    JSON output ``{"result": [ ... ]}`` and XML output ``<app><result>
    ... </result></app>``

    The `kwargs` can supply any configuration variables which are known to
    other parts of the REST API implementation. The arguments eventually
    become fields in the "API object" created by :meth:`.MiniRESTApi._addAPI`,
    and accessible among other things to the validator functions. The most
    commonly used arguments include those listed below. See the rest of
    the module documentation for the details on them.

    =================== ======================================================
    Keyword             Purpose
    =================== ======================================================
    generate            Name of the response wrapper as described above.
    columns             List of column names for row-oriented data.
    expires             Response expire time in seconds.
    expires_opts        Additional "Cache-Control" options.
    formats             "Accept" formats and associated formatter objects.
    etag_limit          Amount of output to buffer for ETag calculation.
    compression         "Accept-Encoding" methods, empty disables compression.
    compression_level   ZLIB compression level for output (0 .. 9).
    compression_chunk   Approximate amount of output to compress at once.
    =================== ======================================================

    :returns: The original function suitably enriched with attributes if
      invoked as a function-like decorator, or a function which will apply
      the decoration if invoked as bare-word style decorator.
    """

    def apply_restcall_opts(func, args=args, generate=generate, kwargs=kwargs):
        if not func:
            raise ValueError("'restcall' must be applied to a function")
        if args == None:
            args = [a for a in inspect.getargspec(func).args if a != 'self']
        if args == None or not isinstance(args, list):
            raise ValueError("'args' must be defined")
        kwargs.update(generate=generate)
        setattr(func, 'rest.exposed', True)
        setattr(func, 'rest.args', args or [])
        setattr(func, 'rest.params', kwargs)
        return func

    return (func and apply_restcall_opts(func)) or apply_restcall_opts


def rows(cursor):
    """Utility function to convert a sequence `cursor` to a generator."""
    for row in cursor:
        yield row


def rxfilter(rx, select, cursor):
    """Utility function to convert a sequence `cursor` to a generator, but
    applying a filtering predicate to select which rows to return.

    The assumption is that `cursor` yields uniform sequence objects ("rows"),
    and the `select` operator can be invoked with ``select(row)`` for each
    retrieved row to return a string. If the value returned matches the
    regular expression `rx`, the original row is included in the generator
    output, otherwise it's skipped.

    :arg re.RegexObject rx: Regular expression to match against, or at least
     any object which supports ``rx.match(value)`` on the value returned by
     the ``select(row)`` operator.
    :arg callable select: An operator which returns the item to filter on,
     given a row of values, typically an :func:`operator.itemgetter`.
    :arg sequence cursor: Input sequence."""

    for row in cursor:
        if rx.match(select(row)):
            yield row



#!/usr/bin/env python
"""
_List_

MySQL implementation of Locations.List
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class List(DBFormatter):
    sql = "select id, site_name, pending_slots, running_slots from wmbs_location order by site_name"
